# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

from __future__ import division, unicode_literals
from functools import reduce

"""
This module implements representations of slabs and surfaces, as well as
algorithms for generating them.
"""

__author__ = "Richard Tran, Wenhao Sun, Zihan Xu, Shyue Ping Ong"
__copyright__ = "Copyright 2014, The Materials Virtual Lab"
__version__ = "0.1"
__maintainer__ = "Shyue Ping Ong"
__email__ = "ongsp@ucsd.edu"
__date__ = "6/10/14"

try:
    # New Py>=3.5 import
    from math import gcd
except ImportError:
    # Deprecated import from Py3.5 onwards.
    from fractions import gcd
import math
import itertools
import logging

import numpy as np
from scipy.spatial.distance import squareform
from scipy.cluster.hierarchy import linkage, fcluster

from monty.fractions import lcm

from pymatgen.core.periodic_table import get_el_sp
from pymatgen.core.structure import Structure
from pymatgen.core.lattice import Lattice
from pymatgen.core.sites import PeriodicSite

from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.util.coord_utils import in_coord_list
from pymatgen.analysis.structure_matcher import StructureMatcher


logger = logging.getLogger(__name__)


class Slab(Structure):
    """
    Subclass of Structure representing a Slab. Implements additional
    attributes pertaining to slabs, but the init method does not
    actually implement any algorithm that creates a slab. This is a
    DUMMY class who's init method only holds information about the
    slab. Also has additional methods that returns other information
    about a slab such as the surface area, normal, and atom adsorption.

    Note that all Slabs have the surface normal oriented in the c-direction.
    This means the lattice vectors a and b are in the surface plane and the c
    vector is out of the surface plane (though not necessary perpendicular to
    the surface.)

    .. attribute:: miller_index

        Miller index of plane parallel to surface.

    .. attribute:: scale_factor

        Final computed scale factor that brings the parent cell to the
        surface cell.

    .. attribute:: shift

        The shift value in Angstrom that indicates how much this
        slab has been shifted.
    """

    def __init__(self, lattice, species, coords, miller_index,
                 oriented_unit_cell, shift, scale_factor,
                 validate_proximity=False, to_unit_cell=False,
                 coords_are_cartesian=False, site_properties=None, energy=None):
        """
        Makes a Slab structure, a structure object with additional information
        and methods pertaining to slabs.

        Args:
            lattice (Lattice/3x3 array): The lattice, either as a
                :class:`pymatgen.core.lattice.Lattice` or
                simply as any 2D array. Each row should correspond to a lattice
                vector. E.g., [[10,0,0], [20,10,0], [0,0,30]] specifies a
                lattice with lattice vectors [10,0,0], [20,10,0] and [0,0,30].
            species ([Specie]): Sequence of species on each site. Can take in
                flexible input, including:

                i.  A sequence of element / specie specified either as string
                    symbols, e.g. ["Li", "Fe2+", "P", ...] or atomic numbers,
                    e.g., (3, 56, ...) or actual Element or Specie objects.

                ii. List of dict of elements/species and occupancies, e.g.,
                    [{"Fe" : 0.5, "Mn":0.5}, ...]. This allows the setup of
                    disordered structures.
            coords (Nx3 array): list of fractional/cartesian coordinates of
                each species.
            miller_index ([h, k, l]): Miller index of plane parallel to
                surface. Note that this is referenced to the input structure. If
                you need this to be based on the conventional cell,
                you should supply the conventional structure.
            oriented_unit_cell (Structure): The oriented_unit_cell from which
                this Slab is created (by scaling in the c-direction).
            shift (float): The shift in the c-direction applied to get the
                termination.
            scale_factor (array): scale_factor Final computed scale factor
                that brings the parent cell to the surface cell.
            validate_proximity (bool): Whether to check if there are sites
                that are less than 0.01 Ang apart. Defaults to False.
            coords_are_cartesian (bool): Set to True if you are providing
                coordinates in cartesian coordinates. Defaults to False.
            site_properties (dict): Properties associated with the sites as a
                dict of sequences, e.g., {"magmom":[5,5,5,5]}. The sequences
                have to be the same length as the atomic species and
                fractional_coords. Defaults to None for no properties.
            energy (float): A value for the energy.
        """
        self.oriented_unit_cell = oriented_unit_cell
        self.miller_index = tuple(miller_index)
        self.shift = shift
        self.scale_factor = scale_factor
        self.energy = energy
        super(Slab, self).__init__(
            lattice, species, coords, validate_proximity=validate_proximity,
            to_unit_cell=to_unit_cell,
            coords_are_cartesian=coords_are_cartesian,
            site_properties=site_properties)

    def get_orthogonal_c_slab(self):
        """
        This method returns a Slab where the normal (c lattice vector) is
        "forced" to be exactly orthogonal to the surface a and b lattice
        vectors. **Note that this breaks inherent symmetries in the slab.**
        It should be pointed out that orthogonality is not required to get good
        surface energies, but it can be useful in cases where the slabs are
        subsequently used for postprocessing of some kind, e.g. generating
        GBs or interfaces.
        """
        a, b, c = self.lattice.matrix
        new_c = np.cross(a, b)
        new_c /= np.linalg.norm(new_c)
        new_c = np.dot(c, new_c) * new_c
        new_latt = Lattice([a, b, new_c])
        return Slab(lattice=new_latt, species=self.species,
                    coords=self.cart_coords, miller_index=self.miller_index,
                    oriented_unit_cell=self.oriented_unit_cell,
                    shift=self.shift, scale_factor=self.scale_factor,
                    coords_are_cartesian=True, energy=self.energy)

    def get_sorted_structure(self, key=None, reverse=False):
        """
        Get a sorted copy of the structure. The parameters have the same
        meaning as in list.sort. By default, sites are sorted by the
        electronegativity of the species. Note that Slab has to override this
        because of the different __init__ args.

        Args:
            key: Specifies a function of one argument that is used to extract
                a comparison key from each list element: key=str.lower. The
                default value is None (compare the elements directly).
            reverse (bool): If set to True, then the list elements are sorted
                as if each comparison were reversed.
        """
        sites = sorted(self, key=key, reverse=reverse)
        s = Structure.from_sites(sites)
        return Slab(s.lattice, s.species_and_occu, s.frac_coords,
                    self.miller_index, self.oriented_unit_cell, self.shift,
                    self.scale_factor, site_properties=s.site_properties)

    def copy(self, site_properties=None, sanitize=False):
        """
        Convenience method to get a copy of the structure, with options to add
        site properties.

        Args:
            site_properties (dict): Properties to add or override. The
                properties are specified in the same way as the constructor,
                i.e., as a dict of the form {property: [values]}. The
                properties should be in the order of the *original* structure
                if you are performing sanitization.
            sanitize (bool): If True, this method will return a sanitized
                structure. Sanitization performs a few things: (i) The sites are
                sorted by electronegativity, (ii) a LLL lattice reduction is
                carried out to obtain a relatively orthogonalized cell,
                (iii) all fractional coords for sites are mapped into the
                unit cell.

        Returns:
            A copy of the Structure, with optionally new site_properties and
            optionally sanitized.
        """
        props = self.site_properties
        if site_properties:
            props.update(site_properties)
        return Slab(self.lattice, self.species_and_occu, self.frac_coords,
                    self.miller_index, self.oriented_unit_cell, self.shift,
                    self.scale_factor, site_properties=props)

    @property
    def dipole(self):
        """
        Calculates the dipole of the Slab in the direction of the surface
        normal. Note that the Slab must be oxidation state-decorated for this
        to work properly. Otherwise, the Slab will always have a dipole of 0.
        """
        dipole = np.zeros(3)
        mid_pt = np.sum(self.cart_coords, axis=0) / len(self)
        normal = self.normal
        for site in self:
            charge = sum([getattr(sp, "oxi_state", 0) * amt
                          for sp, amt in site.species_and_occu.items()])
            dipole += charge * np.dot(site.coords - mid_pt, normal) * normal
        return dipole

    def is_polar(self, tol_dipole_per_unit_area=1e-3):
        """
        Checks whether the surface is polar by computing the dipole per unit
        area. Note that the Slab must be oxidation state-decorated for this
        to work properly. Otherwise, the Slab will always be non-polar.

        Args:
            tol_dipole_per_unit_area (float): A tolerance. If the dipole
                magnitude per unit area is less than this value, the Slab is
                considered non-polar. Defaults to 1e-3, which is usually
                pretty good. Normalized dipole per unit area is used as it is
                more reliable than using the total, which tends to be larger for
                slabs with larger surface areas.
        """
        dip_per_unit_area = self.dipole / self.surface_area
        return np.linalg.norm(dip_per_unit_area) > tol_dipole_per_unit_area

    @property
    def normal(self):
        """
        Calculates the surface normal vector of the slab
        """
        normal = np.cross(self.lattice.matrix[0], self.lattice.matrix[1])
        normal /= np.linalg.norm(normal)
        return normal

    @property
    def surface_area(self):
        """
        Calculates the surface area of the slab
        """
        m = self.lattice.matrix
        return np.linalg.norm(np.cross(m[0], m[1]))

    def add_adsorbate_atom(self, indices, specie, distance):
        """
        Gets the structure of single atom adsorption.
        slab structure from the Slab class(in [0, 0, 1])

        Args:
            indices ([int]): Indices of sites on which to put the absorbate.
                Absorbed atom will be displaced relative to the center of
                these sites.
            specie (Specie/Element/str): adsorbed atom species
            distance (float): between centers of the adsorbed atom and the
                given site in Angstroms.
        """
        #Let's do the work in cartesian coords
        center = np.sum([self[i].coords for i in indices], axis=0) / len(
            indices)

        coords = center + self.normal * distance / np.linalg.norm(self.normal)

        self.append(specie, coords, coords_are_cartesian=True)

    def __str__(self):
        comp = self.composition
        outs = [
            "Slab Summary (%s)" % comp.formula,
            "Reduced Formula: %s" % comp.reduced_formula,
            "Miller index: %s" % (self.miller_index, ),
            "Shift: %.4f, Scale Factor: %s" % (self.shift,
                                               self.scale_factor.__str__())]
        to_s = lambda x: "%0.6f" % x
        outs.append("abc   : " + " ".join([to_s(i).rjust(10)
                                           for i in self.lattice.abc]))
        outs.append("angles: " + " ".join([to_s(i).rjust(10)
                                           for i in self.lattice.angles]))
        outs.append("Sites ({i})".format(i=len(self)))
        for i, site in enumerate(self):
            outs.append(" ".join([str(i + 1), site.species_string,
                                  " ".join([to_s(j).rjust(12)
                                            for j in site.frac_coords])]))
        return "\n".join(outs)

    def as_dict(self):
        d = super(Slab, self).as_dict()
        d["@module"] = self.__class__.__module__
        d["@class"] = self.__class__.__name__
        d["oriented_unit_cell"] = self.oriented_unit_cell.as_dict()
        d["miller_index"] = self.miller_index
        d["shift"] = self.shift
        d["scale_factor"] = self.scale_factor
        d["energy"] = self.energy
        return d

    @classmethod
    def from_dict(cls, d):
        lattice = Lattice.from_dict(d["lattice"])
        sites = [PeriodicSite.from_dict(sd, lattice) for sd in d["sites"]]
        s = Structure.from_sites(sites)

        return Slab(
            lattice=lattice,
            species=s.species_and_occu, coords=s.frac_coords,
            miller_index=d["miller_index"],
            oriented_unit_cell=Structure.from_dict(d["oriented_unit_cell"]),
            shift=d["shift"], scale_factor=d["scale_factor"],
            site_properties=s.site_properties, energy=d["energy"]
        )


class SlabGenerator(object):

    """
    This class generates different slabs using shift values determined by where a
    unique termination can be found along with other criterias such as where a
    termination doesn't break a polyhedral bond. The shift value then indicates
    where the slab layer will begin and terminate in the slab-vacuum system.

    .. attribute:: oriented_unit_cell

        A unit cell of the parent structure with the miller
        index of plane parallel to surface

    .. attribute:: parent

        Parent structure from which Slab was derived.

    .. attribute:: lll_reduce

        Whether or not the slabs will be orthogonalized

    .. attribute:: center_slab

        Whether or not the slabs will be centered between
        the vacuum layer

    .. attribute:: slab_scale_factor

        Final computed scale factor that brings the parent cell to the
        surface cell.

    .. attribute:: miller_index

        Miller index of plane parallel to surface.

    .. attribute:: min_slab_size

        Minimum size in angstroms of layers containing atoms

    .. attribute:: min_vac_size

        Minimize size in angstroms of layers containing vacuum

    """

    def __init__(self, initial_structure, miller_index, min_slab_size,
                 min_vacuum_size, lll_reduce=False, center_slab=False,
                 primitive=True, max_normal_search=None):
        """
        Calculates the slab scale factor and uses it to generate a unit cell
        of the initial structure that has been oriented by its miller index.
        Also stores the initial information needed later on to generate a slab.

        Args:
            initial_structure (Structure): Initial input structure. Note that to
                ensure that the miller indices correspond to usual
                crystallographic definitions, you should supply a conventional
                unit cell structure.
            miller_index ([h, k, l]): Miller index of plane parallel to
                surface. Note that this is referenced to the input structure. If
                you need this to be based on the conventional cell,
                you should supply the conventional structure.
            min_slab_size (float): In Angstroms
            min_vac_size (float): In Angstroms
            lll_reduce (bool): Whether to perform an LLL reduction on the
                eventual structure.
            center_slab (bool): Whether to center the slab in the cell with
                equal vacuum spacing from the top and bottom.
            primitive (bool): Whether to reduce any generated slabs to a
                primitive cell (this does **not** mean the slab is generated
                from a primitive cell, it simply means that after slab
                generation, we attempt to find shorter lattice vectors,
                which lead to less surface area and smaller cells).
            max_normal_search (int): If set to a positive integer, the code will
                conduct a search for a normal lattice vector that is as
                perpendicular to the surface as possible by considering
                multiples linear combinations of lattice vectors up to
                max_normal_search. This has no bearing on surface energies,
                but may be useful as a preliminary step to generating slabs
                for absorption and other sizes. It is typical that this will
                not be the smallest possible cell for simulation. Normality
                is not guaranteed, but the oriented cell will have the c
                vector as normal as possible (within the search range) to the
                surface. A value of up to the max absolute Miller index is
                usually sufficient.
        """
        latt = initial_structure.lattice
        d = abs(reduce(gcd, miller_index))
        miller_index = tuple([int(i / d) for i in miller_index])
        #Calculate the surface normal using the reciprocal lattice vector.
        recp = latt.reciprocal_lattice_crystallographic
        normal = recp.get_cartesian_coords(miller_index)
        normal /= np.linalg.norm(normal)

        slab_scale_factor = []
        non_orth_ind = []
        eye = np.eye(3, dtype=np.int)
        for i, j in enumerate(miller_index):
            if j == 0:
                # Lattice vector is perpendicular to surface normal, i.e.,
                # in plane of surface. We will simply choose this lattice
                # vector as one of the basis vectors.
                slab_scale_factor.append(eye[i])
            else:
                #Calculate projection of lattice vector onto surface normal.
                d = abs(np.dot(normal, latt.matrix[i])) / latt.abc[i]
                non_orth_ind.append((i, d))

        # We want the vector that has maximum magnitude in the
        # direction of the surface normal as the c-direction.
        # Results in a more "orthogonal" unit cell.
        c_index, dist = max(non_orth_ind, key=lambda t: t[1])

        if len(non_orth_ind) > 1:
            lcm_miller = lcm(*[miller_index[i] for i, d in non_orth_ind])
            for (i, di), (j, dj) in itertools.combinations(non_orth_ind, 2):
                l = [0, 0, 0]
                l[i] = -int(round(lcm_miller / miller_index[i]))
                l[j] = int(round(lcm_miller / miller_index[j]))
                slab_scale_factor.append(l)
                if len(slab_scale_factor) == 2:
                    break

        if max_normal_search is None:
            slab_scale_factor.append(eye[c_index])
        else:
            index_range = sorted(
                reversed(range(-max_normal_search, max_normal_search + 1)),
                key=lambda x: abs(x))
            candidates = []
            for uvw in itertools.product(index_range, index_range, index_range):
                if (not any(uvw)) or abs(
                        np.linalg.det(slab_scale_factor + [uvw])) < 1e-8:
                    continue
                vec = latt.get_cartesian_coords(uvw)
                l = np.linalg.norm(vec)
                cosine = abs(np.dot(vec, normal) / l)
                candidates.append((uvw, cosine, l))
                if abs(abs(cosine) - 1) < 1e-8:
                    # If cosine of 1 is found, no need to search further.
                    break
            # We want the indices with the maximum absolute cosine,
            # but smallest possible length.
            uvw, cosine, l = max(candidates, key=lambda x: (x[1], -l))
            slab_scale_factor.append(uvw)

        slab_scale_factor = np.array(slab_scale_factor)

        # Let's make sure we have a left-handed crystallographic system
        if np.linalg.det(slab_scale_factor) < 0:
            slab_scale_factor *= -1

        single = initial_structure.copy()
        single.make_supercell(slab_scale_factor)

        self.oriented_unit_cell = Structure.from_sites(single,
                                                       to_unit_cell=True)
        self.parent = initial_structure
        self.lll_reduce = lll_reduce
        self.center_slab = center_slab
        self.slab_scale_factor = slab_scale_factor
        self.miller_index = miller_index
        self.min_vac_size = min_vacuum_size
        self.min_slab_size = min_slab_size
        self.primitive = primitive
        self._normal = normal
        a, b, c = self.oriented_unit_cell.lattice.matrix
        self._proj_height = abs(np.dot(normal, c))

    def get_slab(self, shift=0, tol=0.1, energy=None):
        """
        This method takes in shift value for the c lattice direction and
        generates a slab based on the given shift. You should rarely use this
        method. Instead, it is used by other generation algorithms to obtain
        all slabs.

        Arg:
            shift (float): A shift value in Angstrom that determines how much a
                slab should be shifted.
            tol (float): Tolerance to determine primitive cell.
            energy (float): An energy to assign to the slab.

        Returns:
            (Slab) A Slab object with a particular shifted oriented unit cell.
        """

        h = self._proj_height
        nlayers_slab = int(math.ceil(self.min_slab_size / h))
        nlayers_vac = int(math.ceil(self.min_vac_size / h))
        nlayers = nlayers_slab + nlayers_vac

        species = self.oriented_unit_cell.species_and_occu
        props = self.oriented_unit_cell.site_properties
        props = {k: v * nlayers_slab for k, v in props.items()}
        frac_coords = self.oriented_unit_cell.frac_coords
        frac_coords = np.array(frac_coords) +\
                      np.array([0, 0, -shift])[None, :]
        frac_coords = frac_coords - np.floor(frac_coords)
        a, b, c = self.oriented_unit_cell.lattice.matrix
        new_lattice = [a, b, nlayers * c]
        frac_coords[:, 2] = frac_coords[:, 2] / nlayers
        all_coords = []
        for i in range(nlayers_slab):
            fcoords = frac_coords.copy()
            fcoords[:, 2] += i / nlayers
            all_coords.extend(fcoords)

        slab = Structure(new_lattice, species * nlayers_slab, all_coords,
                         site_properties=props)

        scale_factor = self.slab_scale_factor
        # Whether or not to orthogonalize the structure
        if self.lll_reduce:
            lll_slab = slab.copy(sanitize=True)
            mapping = lll_slab.lattice.find_mapping(slab.lattice)
            scale_factor = np.dot(mapping[2], scale_factor)
            slab = lll_slab

        # Whether or not to center the slab layer around the vacuum
        if self.center_slab:
            avg_c = np.average([c[2] for c in slab.frac_coords])
            slab.translate_sites(list(range(len(slab))), [0, 0, 0.5 - avg_c])

        if self.primitive:
            prim = slab.get_primitive_structure(tolerance=tol)
            if energy is not None:
                energy = prim.volume / slab.volume * energy
            slab = prim

        return Slab(slab.lattice, slab.species_and_occu,
                    slab.frac_coords, self.miller_index,
                    self.oriented_unit_cell, shift,
                    scale_factor, site_properties=slab.site_properties,
                    energy=energy)

    def _calculate_possible_shifts(self, tol=0.1):
        frac_coords = self.oriented_unit_cell.frac_coords
        n = len(frac_coords)

        if n == 1:
            # Clustering does not work when there is only one data point.
            shift = frac_coords[0][2] + 0.5
            return [shift - math.floor(shift)]

        # We cluster the sites according to the c coordinates. But we need to
        # take into account PBC. Let's compute a fractional c-coordinate
        # distance matrix that accounts for PBC.
        dist_matrix = np.zeros((n, n))
        h = self._proj_height
        # Projection of c lattice vector in
        # direction of surface normal.
        for i, j in itertools.combinations(list(range(n)), 2):
            if i != j:
                cdist = frac_coords[i][2] - frac_coords[j][2]
                cdist = abs(cdist - round(cdist)) * h
                dist_matrix[i, j] = cdist
                dist_matrix[j, i] = cdist

        condensed_m = squareform(dist_matrix)
        z = linkage(condensed_m)
        clusters = fcluster(z, tol, criterion="distance")

        #Generate dict of cluster# to c val - doesn't matter what the c is.
        c_loc = {c: frac_coords[i][2] for i, c in enumerate(clusters)}

        #Put all c into the unit cell.
        possible_c = [c - math.floor(c) for c in sorted(c_loc.values())]

        # Calculate the shifts
        nshifts = len(possible_c)
        shifts = []
        for i in range(nshifts):
            if i == nshifts - 1:
                # There is an additional shift between the first and last c
                # coordinate. But this needs special handling because of PBC.
                shift = (possible_c[0] + 1 + possible_c[i]) * 0.5
                if shift > 1:
                    shift -= 1
            else:
                shift = (possible_c[i] + possible_c[i + 1]) * 0.5
            shifts.append(shift - math.floor(shift))
        shifts = sorted(shifts)
        return shifts

    def _get_c_ranges(self, bonds):
        c_ranges = set()
        bonds = {(get_el_sp(s1), get_el_sp(s2)): dist for (s1, s2), dist in
                 bonds.items()}
        for (sp1, sp2), bond_dist in bonds.items():
            for site in self.oriented_unit_cell:
                if sp1 in site.species_and_occu:
                    for nn, d in self.oriented_unit_cell.get_neighbors(
                            site, bond_dist):
                        if sp2 in nn.species_and_occu:
                            c_range = tuple(sorted([site.frac_coords[2],
                                                    nn.frac_coords[2]]))
                            if c_range[1] > 1:
                                # Takes care of PBC when c coordinate of site
                                # goes beyond the upper boundary of the cell
                                c_ranges.add((c_range[0], 1))
                                c_ranges.add((0, c_range[1] - 1))
                            elif c_range[0] < 0:
                                # Takes care of PBC when c coordinate of site
                                # is below the lower boundary of the unit cell
                                c_ranges.add((0, c_range[1]))
                                c_ranges.add((c_range[0] + 1, 1))
                            elif c_range[0] != c_range[1]:
                                c_ranges.add(c_range)
        return c_ranges

    def get_slabs(self, bonds=None, tol=0.1, max_broken_bonds=0):
        """
        This method returns a list of slabs that are generated using the list of
        shift values from the method, _calculate_possible_shifts(). Before the
        shifts are used to create the slabs however, if the user decides to take
        into account whether or not a termination will break any polyhedral
        structure (bonds != None), this method will filter out any shift values
        that do so.

        Args:
            bonds ({(specie1, specie2): max_bond_dist}: bonds are
                specified as a dict of tuples: float of specie1, specie2
                and the max bonding distance. For example, PO4 groups may be
                defined as {("P", "O"): 3}.
            tol (float): Threshold parameter in fcluster in order to check
                if two atoms are lying on the same plane. Default thresh set
                to 0.1 Angstrom in the direction of the surface normal.
            max_broken_bonds (int): Maximum number of allowable broken bonds
                for the slab. Use this to limit # of slabs (some structures
                may have a lot of slabs). Defaults to zero, which means no
                defined bonds must be broken.

        Returns:
            ([Slab]) List of all possible terminations of a particular surface.
            Slabs are sorted by the # of bonds broken.
        """
        c_ranges = set() if bonds is None else self._get_c_ranges(bonds)

        slabs = []
        for shift in self._calculate_possible_shifts(tol=tol):
            bonds_broken = 0
            for r in c_ranges:
                if r[0] <= shift <= r[1]:
                    bonds_broken += 1
            if bonds_broken <= max_broken_bonds:
                # For now, set the energy to be equal to no. of broken bonds
                # per unit cell.
                slab = self.get_slab(shift, tol=tol, energy=bonds_broken)
                slabs.append(slab)

        # Further filters out any surfaces made that might be the same
        m = StructureMatcher(ltol=tol, stol=tol, primitive_cell=False,
                             scale=False)
        slabs = [g[0] for g in m.group_structures(slabs)]
        return sorted(slabs, key=lambda s: s.energy)


def get_symmetrically_distinct_miller_indices(structure, max_index):
    """
    Returns all symmetrically distinct indices below a certain max-index for
    a given structure. Analysis is based on the symmetry of the reciprocal
    lattice of the structure.

    Args:
        structure (Structure): input structure.
        max_index (int): The maximum index. For example, a max_index of 1
            means that (100), (110), and (111) are returned for the cubic
            structure. All other indices are equivalent to one of these.
    """

    recp_lattice = structure.lattice.reciprocal_lattice_crystallographic
    # Need to make sure recp lattice is big enough, otherwise symmetry
    # determination will fail. We set the overall volume to 1.
    recp_lattice = recp_lattice.scale(1)
    recp = Structure(recp_lattice, ["H"], [[0, 0, 0]])

    # Creates a function that uses the symmetry operations in the
    # structure to find Miller indices that might give repetitive slabs
    analyzer = SpacegroupAnalyzer(recp, symprec=0.001)
    symm_ops = analyzer.get_symmetry_operations()
    unique_millers = []

    def is_already_analyzed(miller_index):
        for op in symm_ops:
            if in_coord_list(unique_millers, op.operate(miller_index)):
                return True
        return False

    r = list(range(-max_index, max_index + 1))
    r.reverse()
    for miller in itertools.product(r, r, r):
        if any([i != 0 for i in miller]):
            d = abs(reduce(gcd, miller))
            miller = tuple([int(i / d) for i in miller])
            if not is_already_analyzed(miller):
                unique_millers.append(miller)
    return unique_millers


def generate_all_slabs(structure, max_index, min_slab_size, min_vacuum_size,
                       bonds=None, tol=1e-3, max_broken_bonds=0,
                       lll_reduce=False, center_slab=False, primitive=True,
                       max_normal_search=None):
    """
    A function that finds all different slabs up to a certain miller index.
    Slabs oriented under certain Miller indices that are equivalent to other
    slabs in other Miller indices are filtered out using symmetry operations
    to get rid of any repetitive slabs. For example, under symmetry operations,
    CsCl has equivalent slabs in the (0,0,1), (0,1,0), and (1,0,0) direction.

    Args:
        structure (Structure): Initial input structure. Note that to
                ensure that the miller indices correspond to usual
                crystallographic definitions, you should supply a conventional
                unit cell structure.
        max_index (int): The maximum Miller index to go up to.
        min_slab_size (float): In Angstroms
        min_vac_size (float): In Angstroms
        bonds ({(specie1, specie2): max_bond_dist}: bonds are
            specified as a dict of tuples: float of specie1, specie2
            and the max bonding distance. For example, PO4 groups may be
            defined as {("P", "O"): 3}.
        tol (float): Threshold parameter in fcluster in order to check
            if two atoms are lying on the same plane. Default thresh set
            to 0.1 Angstrom in the direction of the surface normal.
        max_broken_bonds (int): Maximum number of allowable broken bonds
            for the slab. Use this to limit # of slabs (some structures
            may have a lot of slabs). Defaults to zero, which means no
            defined bonds must be broken.
        lll_reduce (bool): Whether to perform an LLL reduction on the
            eventual structure.
        center_slab (bool): Whether to center the slab in the cell with
            equal vacuum spacing from the top and bottom.
        primitive (bool): Whether to reduce any generated slabs to a
            primitive cell (this does **not** mean the slab is generated
            from a primitive cell, it simply means that after slab
            generation, we attempt to find shorter lattice vectors,
            which lead to less surface area and smaller cells).
        max_normal_search (int): If set to a positive integer, the code will
            conduct a search for a normal lattice vector that is as
            perpendicular to the surface as possible by considering
            multiples linear combinations of lattice vectors up to
            max_normal_search. This has no bearing on surface energies,
            but may be useful as a preliminary step to generating slabs
            for absorption and other sizes. It is typical that this will
            not be the smallest possible cell for simulation. Normality
            is not guaranteed, but the oriented cell will have the c
            vector as normal as possible (within the search range) to the
            surface. A value of up to the max absolute Miller index is
            usually sufficient.
    """
    all_slabs = []
    for miller in get_symmetrically_distinct_miller_indices(
            structure, max_index):
        gen = SlabGenerator(structure.copy(), miller, min_slab_size,
                            min_vacuum_size, lll_reduce=lll_reduce,
                            center_slab=center_slab, primitive=primitive,
                            max_normal_search=max_normal_search)
        slabs = gen.get_slabs(bonds=bonds, tol=tol,
                              max_broken_bonds=max_broken_bonds)
        if len(slabs) > 0:
            logger.debug("%s has %d slabs... " % (miller, len(slabs)))
            all_slabs.extend(slabs)

    return all_slabs



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

from __future__ import division, print_function, unicode_literals
from __future__ import absolute_import

"""
This module provides a base class for tensor-like objects and methods for
basic tensor manipulation.  It also provides a class, SquareTensor,
that provides basic methods for creating and manipulating rank 2 tensors
"""


__author__ = "Maarten de Jong"
__copyright__ = "Copyright 2012, The Materials Project"
__credits__ = ("Joseph Montoya, Shyam Dwaraknath, Wei Chen, "
               "Mark Asta, Anubhav Jain, Terence Lew")
__version__ = "1.0"
__maintainer__ = "Joseph Montoya"
__email__ = "montoyjh@lbl.gov"
__status__ = "Development"
__date__ = "March 22, 2012"


from scipy.linalg import polar
import numpy as np
import itertools
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.core.operations import SymmOp

class TensorBase(np.ndarray):
    """
    Base class for doing useful general operations on Nth order tensors,
    without restrictions on the type (stress, elastic, strain, piezo, etc.)
    """

    def __new__(cls, input_array):
        """
        Create a TensorBase object.  Note that the constructor uses __new__
        rather than __init__ according to the standard method of
        subclassing numpy ndarrays.

        Args:
            input_array: (3xN array-like): the 3xN array-like representing
                a tensor quantity
        """
        obj = np.asarray(input_array).view(cls)
        obj.rank = len(obj.shape)
        if not all([i == 3 for i in obj.shape]):
            raise ValueError("Pymatgen only supports 3-dimensional tensors")
        
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.rank = getattr(obj, 'rank', None)

    def __array_wrap__(self, obj):
        """
        Overrides __array_wrap__ methods in ndarray superclass to avoid errors
        associated with functions that return scalar values
        """

        if len(obj.shape) == 0:
            return obj[()]
        else:
            return np.ndarray.__array_wrap__(self, obj)
        
    def __hash__(self):
        """
        define a hash function, since numpy arrays
        have their own __eq__ method
        """
        return hash(self.tostring())

    def __repr__(cls):
        return "{}({})".format(cls.__class__.__name__,
                               cls.__str__())

    def zeroed(self, tol = 1e-3):
        """
        returns the matrix with all entries below a certain threshold
        (i.e. tol) set to zero
        """
        new_tensor = self.copy()
        new_tensor[abs(new_tensor) < tol] = 0
        return new_tensor

    def transform(self, symm_op):
        """
        Applies a transformation (via a symmetry operation) to a tensor. 

        Args:
            symm_op (3x3 array-like): a symmetry operation to apply to the tensor
        """

        return symm_op.transform_tensor(self)

    def rotate(self, matrix, tol=1e-5):
        matrix = SquareTensor(matrix)
        if not matrix.is_rotation(tol):
            raise ValueError("Rotation matrix is not valid.")
        sop = SymmOp.from_rotation_and_translation(matrix,
                                                   [0., 0., 0.])
        return self.transform(sop)


    @property
    def symmetrized(self):
        """
        Returns a generally symmetrized tensor, calculated by taking 
        the sum of the tensor and its transpose with respect to all 
        possible permutations of indices
        """
        perms = list(itertools.permutations(range(self.rank)))
        return sum([np.transpose(self, ind) for ind in perms]) / len(perms)

    def is_symmetric(self, tol=1e-5):
        """
        Tests whether a tensor is symmetric or not based on the residual
        with its symmetric part, from self.symmetrized

        Args:
            tol (float): tolerance to test for symmetry
        """
        return (self - self.symmetrized < tol).all()

    def fit_to_structure(self, structure, symprec = 0.1):
        """
        Returns a tensor that is invariant with respect to symmetry
        operations corresponding to a structure

        Args: 
            structure (Structure): structure from which to generate 
                symmetry operations
            symprec (float): symmetry tolerance for the Spacegroup Analyzer
                used to generate the symmetry operations
        """
        sga = SpacegroupAnalyzer(structure, symprec)
        symm_ops = sga.get_symmetry_operations(cartesian=True)
        return sum([self.transform(symm_op)
                    for symm_op in symm_ops]) / len(symm_ops)

    def is_fit_to_structure(self, structure, tol=1e-2):
        """
        Tests whether a tensor is invariant with respect to the
        symmetry operations of a particular structure by testing
        whether the residual of the symmetric portion is below a 
        tolerance
        
        Args:
            tol (float): tolerance for symmetry testing
        """

        return (self - self.fit_to_structure(structure) < tol).all()


class SquareTensor(TensorBase):
    """
    Base class for doing useful general operations on second rank tensors
    (stress, strain etc.).
    """

    def __new__(cls, input_array):
        """
        Create a SquareTensor object.  Note that the constructor uses __new__
        rather than __init__ according to the standard method of
        subclassing numpy ndarrays.  Error is thrown when the class is
        initialized with non-square matrix.

        Args:
            stress_matrix (3x3 array-like): the 3x3 array-like
                representing the Green-Lagrange strain
        """

        obj = TensorBase(input_array).view(cls)
        if not (len(obj.shape) == 2):
            raise ValueError("SquareTensor only takes 2-D "
                             "tensors as input")
        return obj
        
    @property
    def trans(self):
        """
        shorthand for transpose on SquareTensor
        """
        return SquareTensor(np.transpose(self))

    @property
    def inv(self):
        """
        shorthand for matrix inverse on SquareTensor
        """
        if self.det == 0:
            raise ValueError("SquareTensor is non-invertible")
        return SquareTensor(np.linalg.inv(self))

    @property
    def det(self):
        """
        shorthand for the determinant of the SquareTensor
        """
        return np.linalg.det(self)

    def is_rotation(self, tol=1e-5):
        """
        Test to see if tensor is a valid rotation matrix, performs a
        test to check whether the inverse is equal to the transpose
        and if the determinant is equal to one within the specified
        tolerance

        Args:
            tol (float): tolerance to both tests of whether the
                the determinant is one and the inverse is equal
                to the transpose
        """

        return (np.abs(self.inv - self.trans) < tol).all() \
            and (np.linalg.det(self) - 1. < tol)

    def get_scaled(self, scale_factor):
        """
        Scales the tensor by a certain multiplicative scale factor

        Args:
            scale_factor (float): scalar multiplier to be applied to the
                SquareTensor object
        """
        return SquareTensor(self * scale_factor)

    @property
    def principal_invariants(self):
        """
        Returns a list of principal invariants for the tensor,
        which are the values of the coefficients of the characteristic
        polynomial for the matrix
        """
        return np.poly(self)[1:]*np.array([-1, 1, -1])

    def polar_decomposition(self, side='right'):
        """
        calculates matrices for polar decomposition
        """
        return polar(self, side=side)



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

from __future__ import division, print_function, unicode_literals
from __future__ import absolute_import

"""
This module provides classes for the Piezoelectric tensor
"""
from pymatgen.core.operations import SymmOp
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
import numpy as np
import warnings
from six.moves import range

__author__ = "Shyam Dwaraknath"
__copyright__ = "Copyright 2016, The Materials Project"
__version__ = "1.0"
__maintainer__ = "Shyam Dwaraknath"
__email__ = "shyamd@lbl.gov"
__status__ = "Development"
__date__ = "Feb, 2016"

voigt_map = [(0, 0), (1, 1), (2, 2), (1, 2), (2, 0), (0, 1)]


class PiezoTensor(np.ndarray):
    """
    This class describes the 3 x 6 piezo tensor in Voigt-notation
    """

    def __new__(cls, input_array):
        """
        Create an PiezoTensor object.  The constructor throws an error if
        the shape of the input_matrix argument is not 3x6, i. e. in Voigt-
        notation. Note that the constructor uses __new__ rather than
        __init__ according to the standard method of subclassing numpy
        ndarrays.

        Args:
            input_matrix (3x6 array-like): the Voigt-notation 3x6 array-like
                representing the piezo tensor
        """
        obj = np.asarray(input_array).view(cls)
        if obj.shape != (3, 6):
            raise ValueError("Default piezo tensor constructor requires "
                             "input argument to be the Voigt-notation 3x6 "
                             "array.  To construct from a 3x3x3 array, use "
                             "PiezoTensor.from_full_tensor")
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return

    @property
    def full_tensor(self):
        """
        Returns the tensor in standard notation (i. e.
        a 3th order 3-dimensional tensor, C_{ijk}), which
        is represented in a np.array with shape (3,3,3)
        """
        c = np.zeros((3, 3, 3))
        for i in range(3):
            for p in range(6):
                j, k = voigt_map[p]
                c[i, j, k] = c[i, k, j] = self[i, p]
        return c

    @classmethod
    def from_full_tensor(cls, c_ijk, tol=1e-5):
        """
        Factory method to construct piezo tensor from third order
        tensor C_ijk.  First tests for appropriate symmetries and then
        constructs the 3x6 voigt notation tensor.

        Args:
            c_ijkl (3x3x3 array-like): third order tensor corresponding
                to the piezo tensor
            tol (float): tolerance for the symmetry test of the tensor
        """
        # Test symmetry of elastic tensor
        c_ijk = np.array(c_ijk)
        if not (c_ijk - np.transpose(c_ijk, (0, 2, 1)) < tol).all():
            raise ValueError("Input piezo tensor does "
                             "not satisfy necessary symmetries")
        # Construct elastic tensor
        c_ip = np.zeros((3, 6))
        for i in range(3):
            for p in range(6):
                j, k = voigt_map[p]
                c_ip[i, p] = c_ijk[i, j, k]

        return cls(c_ip)

    def __array_wrap__(self, obj):
        """
        Overrides __array_wrap__ methods in ndarray superclass to avoid errors
        associated with functions that return scalar values
        """

        if len(obj.shape) == 0:
            return obj[()]
        else:
            return np.ndarray.__array_wrap__(self, obj)

    def __hash__(self):
        """
        define a hash function, since numpy arrays
        have their own __eq__ method
        """
        return hash(self.tostring())

    def __repr__(cls):
        return "{}({})".format(cls.__class__.__name__,
                               cls.__str__())

    def transform(self, sym):
        """
        Returns a transformed tensor based on input of symmetry operation

        Args:
            symp (SymmOp): symmetry operation
        """
        return PiezoTensor.from_full_tensor(sym.transform_tensor(self.full_tensor))

    def is_valid(self, structure, symprec=0.1, tol=1e-3):
        """
        Checks the piezo tensor against the symmetries of the structure and
        determine if the piezoelectric tensor is valid for that structure

        Args:
            structure (Structure): structure to check against
            symprec (float): Tolerance for symmetry finding, default good for
                             as-made structures. Set to 0.1 for MP structures
            tol (float): tolerance for equality
        """

        sg = SpacegroupAnalyzer(structure, symprec)

        # Check every symmetry operation in the space group
        for symm in sg.get_symmetry_operations(cartesian=True):
            # does it produce the same tensor?
            diff = self.transform(symm) - self
            if not (np.abs(diff) < tol).all():
                print(diff)
                return False

        return True

    def symmeterize(self, structure, symprec=0.1, tol=1e-3):
        """
        Averages the piezo tensor based on the symmetries of the structure

        Args:
            structure (Structure): structure to check against
            symprec (float): Tolerance for symmetry finding, default good for
                             as-made structures. Set to 0.1 for MP structures
            tol (float): tolerance for zero'ing out indicies. The average procedure produces very small numbers rather then 0. This tolerance is used to zero out those values to make the tensor less messy.
        """

        sg = SpacegroupAnalyzer(structure, symprec)
        new_pt = PiezoTensor(self)

        for symm in sg.get_symmetry_operations(cartesian=True):
            new_pt = (new_pt + new_pt.transform(symm)) / 2

        low_values_indices = np.abs(new_pt) < tol
        new_pt [low_values_indices] = 0

        return new_pt



# coding: utf-8

from __future__ import division, unicode_literals, print_function

"""
This module provides classes to run and analyze boltztrap on pymatgen band
structure objects. Boltztrap is a software interpolating band structures and
computing materials properties from this band structure using Boltzmann
semi-classical transport theory.

Boltztrap has been developped by Georg Madsen.

http://www.icams.de/content/departments/ams/madsen/boltztrap.html

You need the version 1.2.3

References are::

    Madsen, G. K. H., and Singh, D. J. (2006).
    BoltzTraP. A code for calculating band-structure dependent quantities.
    Computer Physics Communications, 175, 67-71
"""

__author__ = "Geoffroy Hautier, Zachary Gibbs"
__copyright__ = "Copyright 2013, The Materials Project"
__version__ = "1.1"
__maintainer__ = "Geoffroy Hautier"
__email__ = "geoffroy@uclouvain.be"
__status__ = "Development"
__date__ = "August 23, 2013"

import os
import math
import numpy as np
import tempfile
from pymatgen.core.lattice import Lattice
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.symmetry.bandstructure import HighSymmKpath
from pymatgen.electronic_structure.dos import Dos, Spin, CompleteDos
from pymatgen.electronic_structure.core import Orbital
from pymatgen.electronic_structure.plotter import DosPlotter
from pymatgen.electronic_structure.bandstructure import BandStructureSymmLine, Kpoint
from monty.os.path import which
from monty.os import cd
from monty.dev import requires
from monty.json import jsanitize
from pymatgen.core.units import Energy, Length
import subprocess

from scipy.constants import e, m_e
from scipy.spatial import distance

try:
    import matplotlib.pyplot as plt
except ImportError:
    pass


class BoltztrapRunner(object):
    """
    This class is used to run Boltztrap on a band structure object.

    Args:
            bs:
                A band structure object
            nelec:
                the number of electrons
            dos_type:
                two options here for the band structure integration: "HISTO"
                (histogram) or "TETRA" using the tetrahedon method. TETRA
                gives typically better results especially for DOSes but takes
                more time
            energy_grid:
                the energy steps used for the integration. in eV
            lpfac:
                the number of interpolation points in the real space. By
                default 10 gives 10 time more points in the real space than
                the number of kpoints given in reciprocal space
            run_type:
                type of boltztrap usage by default BOLTZ to compute transport
                coefficients but you can have also:
                - BANDS to interpolate all bands contained in the energy range specified in 
                energy_span_around_fermi variable, along specified k-points
                - DOS to compute total and partial dos (custom BoltzTraP code needed!)
                - FERMI to compute fermi surface or more correctly to get certain bands interpolated
            band_nb:
                indicates a band number. Used for Fermi Surface interpolation (run_type="FERMI")
            spin:
                specific spin component (1: up, -1: down) of the band selected in FERMI mode (mandatory).
            cond_band:
                if a conduction band is specified in FERMI mode set this variable as True
            tauref:
                reference relaxation time. Only set to a value different than
                zero if we want to model
                beyond the constant relaxation time.
            tauexp:
                exponent for the energy in the non-constant relaxation time
                approach
            tauen:
                reference energy for the non-constant relaxation time approach
            soc:
                results from spin-orbit coupling (soc) computations give
                typically non-polarized (no spin up or down)
                results but 1 electron occupations. If the band structure
                comes from a soc computation, you should set
                soc to True (default False)
            doping:
                the fixed doping levels you want to compute. Boltztrap provides both transport values
                depending on electron chemical potential (fermi energy) and for a series of fixed
                carrier concentrations. By default, this is set to 1e16, 1e17, 1e18, 1e19, 1e20 and 1e21
            energy_span_around_fermi:
                usually the interpolation is not needed on the entire energy range but on a specific range around
                the fermi level. This energy gives this range in eV. by default it is 1.5 eV.
                If DOS or BANDS type are selected, this range is automatically set to cover the entire energy range.
            scissor:
                scissor to the band gap in eV. This applies a scissor operation moving the band edges without changing
                the band shape. This is useful to correct the often underestimated band gap in DFT. Default is 0.0 (no
                scissor)
            kpt_line:
                list/array of kpoints in fractional coordinates for BANDS mode calculation 
                (standard path of high symmetry k-points is automatically set as default)

    """

    @requires(which('x_trans'),
              "BoltztrapRunner requires the executables 'x_trans' to be in "
              "the path. Please download the Boltztrap at "
              "http://www.icams.de/content/departments/ams/madsen/boltztrap"
              ".html and follow the instructions in the README to compile "
              "Bolztrap accordingly. Then add x_trans to your path")
    def __init__(self, bs, nelec, dos_type="HISTO", energy_grid=0.005,
                 lpfac=10, run_type="BOLTZ", band_nb=None, tauref=0, tauexp=0, tauen=0, soc=False, doping=None,
                 energy_span_around_fermi=1.5, scissor=0.0, kpt_line=None, spin=None, cond_band=False):
        self.lpfac = lpfac
        self._bs = bs
        self._nelec = nelec
        self.dos_type = dos_type
        self.energy_grid = energy_grid
        self.error = []
        self.run_type = run_type
        self.band_nb = band_nb
        self.spin = spin
        self.cond_band = cond_band
        self.tauref = tauref
        self.tauexp = tauexp
        self.tauen = tauen
        self.soc = soc
        self.kpt_line = kpt_line
        self.doping = doping or [1e16, 1e17, 1e18, 1e19, 1e20, 1e21]
        self.energy_span_around_fermi = energy_span_around_fermi
        self.scissor = scissor

    def _make_energy_file(self, file_name):
        with open(file_name, 'w') as f:
            f.write("test\n")
            f.write(str(len(self._bs.kpoints)) + "\n")
            sign = 1.0
            if self.cond_band:
                sign = -1.0

            if self.run_type in ("DOS", "BANDS"):
                const = Energy(2, "eV").to("Ry")
                emin_up = np.min(np.array(self._bs._bands[Spin.up]))
                emax_up = np.max(np.array(self._bs._bands[Spin.up]))

                if self._bs.is_spin_polarized:
                    emin_dw = np.min(np.array(self._bs._bands[Spin.down]))
                    low_en_lim = Energy(min((emin_up, emin_dw)) -
                                        self._bs.efermi, "eV").to("Ry")

                    emax_dw = np.max(np.array(self._bs._bands[Spin.down]))
                    high_en_lim = Energy(max((emax_up, emax_dw)) -
                                         self._bs.efermi, "eV").to("Ry")

                    self._ll = low_en_lim - const
                    self._hl = high_en_lim + const

                    en_range = Energy(max((abs(self._ll), abs(self._hl))), "Ry").to("eV")
                else:
                    en_range = Energy(max((abs(emin_up - const), abs(emax_up + const))), "Ry").to("eV")

                self.energy_span_around_fermi = en_range * 1.01
                print("energy_span_around_fermi = ", self.energy_span_around_fermi)

            if self.run_type != "FERMI":
                for i in range(len(self._bs.kpoints)):
                    tmp_eigs = []
                    if self.run_type == "DOS":
                        spin_lst = [self.spin]
                    else:
                        spin_lst = self._bs._bands

                    for spin in spin_lst:
                        for j in range(int(math.floor(self._bs._nb_bands * 0.9))):
                            tmp_eigs.append(Energy(self._bs._bands[Spin(spin)][j][i] -
                                                   self._bs.efermi, "eV").to("Ry"))
                    tmp_eigs.sort()

                    if self.run_type == "DOS" and self._bs.is_spin_polarized:
                        tmp_eigs.insert(0, self._ll)
                        tmp_eigs.append(self._hl)

                    f.write("%12.8f %12.8f %12.8f %d\n"
                            % (self._bs.kpoints[i].frac_coords[0],
                               self._bs.kpoints[i].frac_coords[1],
                               self._bs.kpoints[i].frac_coords[2], len(tmp_eigs)))

                    for j in range(len(tmp_eigs)):
                        f.write("%18.8f\n" % (sign * float(tmp_eigs[j])))

            else:
                for i in range(len(self._bs.kpoints)):
                    tmp_eigs = []
                    tmp_eigs.append(Energy(self._bs._bands[Spin(self.spin)][self.band_nb][i] -
                                           self._bs.efermi, "eV").to("Ry"))
                    f.write("%12.8f %12.8f %12.8f %d\n"
                            % (self._bs.kpoints[i].frac_coords[0],
                               self._bs.kpoints[i].frac_coords[1],
                               self._bs.kpoints[i].frac_coords[2], len(tmp_eigs)))
                    for j in range(len(tmp_eigs)):
                        f.write("%18.8f\n" % (sign * float(tmp_eigs[j])))

    def _make_struc_file(self, file_name):
        sym = SpacegroupAnalyzer(self._bs._structure, symprec=0.01)
        with open(file_name, 'w') as f:
            f.write(self._bs._structure.composition.formula + " " +
                    str(sym.get_spacegroup_symbol()) + "\n")
            for i in range(3):
                line = ''
                for j in range(3):
                    line += "%12.5f" % (
                        Length(self._bs._structure.lattice.matrix[i][j],
                               "ang").to("bohr"))
                f.write(line + '\n')
            ops = sym.get_symmetry_dataset()['rotations']
            f.write(str(len(ops)) + "\n")
            for c in ops:
                f.write('\n'.join([' '.join([str(int(i)) for i in row])
                                   for row in c]))
                f.write('\n')

    # This function is useless in std version of BoltzTraP code
    # because x_trans script overwrite BoltzTraP.def
    def _make_def_file(self, def_file_name):
        with open(def_file_name, 'w') as f:
            so = ""
            if self._bs.is_spin_polarized or self.soc:
                so = "so"
            f.write("5, 'boltztrap.intrans',      'old',    'formatted',0\n" +
                    "6,'boltztrap.outputtrans',      'unknown',    "
                    "'formatted',0\n" +
                    "20,'boltztrap.struct',         'old',    'formatted',0\n" +
                    "10,'boltztrap.energy" + so + "',         'old',    "
                                                  "'formatted',0\n" +
                    "48,'boltztrap.engre',         'unknown',    "
                    "'unformatted',0\n" +
                    "49,'boltztrap.transdos',        'unknown',    "
                    "'formatted',0\n" +
                    "50,'boltztrap.sigxx',        'unknown',    'formatted',"
                    "0\n" +
                    "51,'boltztrap.sigxxx',        'unknown',    'formatted',"
                    "0\n" +
                    "21,'boltztrap.trace',           'unknown',    "
                    "'formatted',0\n" +
                    "22,'boltztrap.condtens',           'unknown',    "
                    "'formatted',0\n" +
                    "24,'boltztrap.halltens',           'unknown',    "
                    "'formatted',0\n" +
                    "30,'boltztrap_BZ.cube',           'unknown',    "
                    "'formatted',0\n")

    # This function is useless in std version of BoltzTraP code
    # because x_trans script overwrite BoltzTraP.def
    def _make_proj_files(self, file_name, def_file_name):
        for o in Orbital:
            for site_nb in range(0, len(self._bs._structure.sites)):
                if o in self._bs._projections[Spin.up][0][0]:
                    with open(file_name + "_" + str(site_nb) + "_" + str(o), 'w') as f:
                        f.write(self._bs._structure.composition.formula + "\n")
                        f.write(str(len(self._bs.kpoints)) + "\n")
                        for i in range(len(self._bs.kpoints)):
                            tmp_proj = []
                            for j in range(int(math.floor(self._bs._nb_bands * 0.9))):
                                tmp_proj.append(self._bs._projections[Spin(self.spin)][j][i][o][site_nb])
                            # TODO deal with the sorting going on at the energy level!!!
                            # tmp_proj.sort()

                            if self.run_type == "DOS" and self._bs.is_spin_polarized:
                                tmp_proj.insert(0, self._ll)
                                tmp_proj.append(self._hl)

                            f.write("%12.8f %12.8f %12.8f %d\n"
                                    % (self._bs.kpoints[i].frac_coords[0],
                                       self._bs.kpoints[i].frac_coords[1],
                                       self._bs.kpoints[i].frac_coords[2], len(tmp_proj)))
                            for j in range(len(tmp_proj)):
                                f.write("%18.8f\n" % float(tmp_proj[j]))
        with open(def_file_name, 'w') as f:
            so = ""
            if self._bs.is_spin_polarized:
                so = "so"
            f.write("5, 'boltztrap.intrans',      'old',    'formatted',0\n" +
                    "6,'boltztrap.outputtrans',      'unknown',    "
                    "'formatted',0\n" +
                    "20,'boltztrap.struct',         'old',    'formatted',0\n" +
                    "10,'boltztrap.energy" + so + "',         'old',    "
                                                  "'formatted',0\n" +
                    "48,'boltztrap.engre',         'unknown',    "
                    "'unformatted',0\n" +
                    "49,'boltztrap.transdos',        'unknown',    "
                    "'formatted',0\n" +
                    "50,'boltztrap.sigxx',        'unknown',    'formatted',"
                    "0\n" +
                    "51,'boltztrap.sigxxx',        'unknown',    'formatted',"
                    "0\n" +
                    "21,'boltztrap.trace',           'unknown',    "
                    "'formatted',0\n" +
                    "22,'boltztrap.condtens',           'unknown',    "
                    "'formatted',0\n" +
                    "24,'boltztrap.halltens',           'unknown',    "
                    "'formatted',0\n" +
                    "30,'boltztrap_BZ.cube',           'unknown',    "
                    "'formatted',0\n")
            i = 1000
            for o in Orbital:
                for site_nb in range(0, len(self._bs._structure.sites)):
                    if o in self._bs._projections[Spin.up][0][0]:
                        f.write(str(i) + ",\'" + "boltztrap.proj_" + str(site_nb) + "_" + str(o.name) +
                                "\' \'old\', \'formatted\',0\n")
                        i += 1

    def _make_intrans_file(self, file_name):
        if self.run_type == "BOLTZ" or self.run_type == "DOS":
            with open(file_name, 'w') as fout:
                fout.write("GENE          # use generic interface\n")
                setgap = 0
                if self.scissor > 0.0001:
                    setgap = 1
                fout.write("1 0 %d %f         # iskip (not presently used) idebug setgap shiftgap \n"
                           % (setgap, Energy(self.scissor, "eV").to("Ry")))
                fout.write(
                    "0.0 %f %f %6.1f     # Fermilevel (Ry),energygrid,energy span around Fermilevel, "
                    "number of electrons\n"
                    % (Energy(self.energy_grid, "eV").to("Ry"), Energy(self.energy_span_around_fermi, "eV").to("Ry"),
                       self._nelec))
                fout.write("CALC                    # CALC (calculate expansion coeff), NOCALC read from file\n")
                fout.write("%d                        # lpfac, number of latt-points per k-point\n" % self.lpfac)
                fout.write("%s                     # run mode (only BOLTZ is supported)\n" % self.run_type)
                fout.write(".15                       # (efcut) energy range of chemical potential\n")
                fout.write("1300. 100.                  # Tmax, temperature grid\n")
                fout.write("-1.  # energyrange of bands given DOS output sig_xxx and dos_xxx (xxx is band number)\n")
                fout.write(self.dos_type + "\n")
                fout.write(str(self.tauref) + " " + str(self.tauexp) + " " + str(self.tauen) + " 0 0 0\n")
                fout.write(str(2 * len(self.doping)) + "\n")
                for d in self.doping:
                    fout.write(str(d) + "\n")
                for d in self.doping:
                    fout.write(str(-d) + "\n")

        elif self.run_type == "FERMI":
            with open(file_name, 'w') as fout:
                fout.write("GENE          # use generic interface\n")
                fout.write(
                    "1 0 0 0.0         # iskip (not presently used) idebug "
                    "setgap shiftgap \n")
                fout.write(
                    "0.0 %f 0.1 %6.1f     # Fermilevel (Ry),energygrid,"
                    "energy span around Fermilevel, "
                    "number of electrons\n"
                    % (Energy(self.energy_grid, "eV").to("Ry"), self._nelec))
                fout.write(
                    "CALC                    # CALC (calculate expansion "
                    "coeff), NOCALC read from file\n")
                fout.write(
                    "%d                        # lpfac, number of latt-points "
                    "per k-point\n" % self.lpfac)
                fout.write(
                    "FERMI                     # run mode (only BOLTZ is "
                    "supported)\n")
                fout.write(str(1) + "                        # actual band selected: " + str(
                    self.band_nb + 1) + " spin: " + str(self.spin))

        elif self.run_type == "BANDS":

            setgap = 0
            if self.scissor > 0.0001:
                setgap = 1
            if self.kpt_line is None:
                kpath = HighSymmKpath(self._bs._structure)
                self.kpt_line = [Kpoint(k, self._bs._structure.lattice) for k in
                                 kpath.get_kpoints(coords_are_cartesian=False)[0]]
                self.kpt_line = np.array([kp.frac_coords for kp in self.kpt_line])

            with open(file_name, 'w') as fout:
                fout.write("GENE          # use generic interface\n")
                fout.write("1 0 %d %f         # iskip (not presently used) idebug setgap shiftgap \n"
                           % (setgap, Energy(self.scissor, "eV").to("Ry")))
                fout.write(
                    "0.0 %f %f %6.1f     # Fermilevel (Ry),energygrid,energy span around Fermilevel, "
                    "number of electrons\n"
                    % (Energy(self.energy_grid, "eV").to("Ry"), Energy(self.energy_span_around_fermi, "eV").to("Ry"),
                       self._nelec))
                fout.write("CALC                    # CALC (calculate expansion coeff), NOCALC read from file\n")
                fout.write("%d                        # lpfac, number of latt-points per k-point\n" % self.lpfac)
                fout.write("BANDS                     # run mode (only BOLTZ is supported)\n")
                fout.write("P " + str(len(self.kpt_line)) + "\n")
                for kp in self.kpt_line:
                    fout.writelines([str(k) + ' ' for k in kp])
                    fout.write('\n')

    def _make_all_files(self, path):
        if self._bs.is_spin_polarized or self.soc:
            self._make_energy_file(os.path.join(path, "boltztrap.energyso"))
        else:
            self._make_energy_file(os.path.join(path, "boltztrap.energy"))

        self._make_struc_file(os.path.join(path, "boltztrap.struct"))
        self._make_intrans_file(os.path.join(path, "boltztrap.intrans"))
        self._make_def_file("BoltzTraP.def")
        if len(self._bs._projections) != 0 and self.run_type == "DOS":
            self._make_proj_files(os.path.join(path, "boltztrap.proj"),
                                  os.path.join(path, "BoltzTraP.def"))

    def run(self, path_dir=None, prev_sigma=None, convergence=True):
        if self.run_type in ("BANDS", "DOS", "FERMI"):
            convergence = False

        if self.run_type == "BANDS" and self._bs.is_spin_polarized:
            print("Reminder: for run_type " + str(self.run_type) + " spin component are not separated!")

        if self.run_type in ("FERMI", "DOS") and self.spin is None:
            if self._bs.is_spin_polarized:
                raise BoltztrapError("Spin component must be specified for spin polarized case!")
            else:
                self.spin = 1

        dir_bz_name = "boltztrap"
        path_dir_orig = path_dir
        if path_dir is None:
            temp_dir = tempfile.mkdtemp()
            path_dir_orig = temp_dir
            path_dir = os.path.join(temp_dir, dir_bz_name)
        else:
            path_dir = os.path.abspath(os.path.join(path_dir_orig, dir_bz_name))
        if not os.path.exists(path_dir):
            os.mkdir(path_dir)
        else:
            for c in os.listdir(path_dir):
                os.remove(path_dir + "/" + c)

        with cd(path_dir):

            ######## convergence loop over energy_grid, lpfac and not on eff_mass (as previously) ########################
            lpfac_start = self.lpfac
            converged = False

            while self.energy_grid > 0.00004:
                sigma_ratio = 1
                self.lpfac = lpfac_start

                print("lpfac, energy_grid: ", self.lpfac, self.energy_grid)

                while self.lpfac < 160:

                    self._make_all_files(path_dir)
                    if self._bs.is_spin_polarized or self.soc:
                        p = subprocess.Popen(["x_trans", "BoltzTraP", "-so"],
                                             stdout=subprocess.PIPE,
                                             stdin=subprocess.PIPE, stderr=subprocess.PIPE)
                        p.wait()
                    else:
                        p = subprocess.Popen(["x_trans", "BoltzTraP"],
                                             stdout=subprocess.PIPE, stdin=subprocess.PIPE,
                                             stderr=subprocess.PIPE)
                        p.wait()

                    for c in p.communicate():
                        print(c)
                        if "STOP error in factorization" in c:
                            raise BoltztrapError("STOP error in factorization")

                    with open(os.path.join(path_dir, dir_bz_name + ".outputtrans")) as f:
                        warning = False
                        for l in f:
                            if "WARNING" in l:
                                warning = True
                                break
                            if "Option unknown" in l:
                                raise BoltztrapError("DOS mode needs a custom version of BoltzTraP code is needed")
                    if warning:
                        print("There was a warning! Increase lpfac to " +
                              str(self.lpfac + 10))
                        self.lpfac += 10
                        continue

                    if convergence:
                        analyzer = BoltztrapAnalyzer.from_files(path_dir)
                        doping_ok = True
                        for doping in ['n', 'p']:
                            for c in analyzer.mu_doping[doping]:
                                if len(analyzer.mu_doping[doping][c]) != len(analyzer.doping[doping]):
                                    doping_ok = False
                                    break
                                if doping == 'p' and \
                                                sorted(analyzer.mu_doping[doping][c], reverse=True) != \
                                                analyzer.mu_doping[doping][c]:
                                    doping_ok = False
                                    break
                                if doping == 'n' and sorted(analyzer.mu_doping[doping][c]) != \
                                        analyzer.mu_doping[doping][c]:
                                    doping_ok = False
                                    break

                        print('doping_ok', doping_ok)
                        if not doping_ok:
                            self.lpfac += 10
                            print("doping not ok, increase lpfac to " + str(self.lpfac))
                            continue

                        converged = True
                        break

                    else:
                        converged = True
                        break

                if not converged:
                    self.energy_grid /= 10
                else:
                    break

            if not converged:
                raise BoltztrapError("Doping convergence not reached with lpfac=" + str(self.lpfac)
                                     + ", energy_grid=" + str(self.energy_grid))

            return path_dir


class BoltztrapError(Exception):
    """
    Exception class for boltztrap.
    Raised when the boltztrap gives an error
    """

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return "BoltztrapError : " + self.msg


class BoltztrapAnalyzer(object):
    """
    Class used to store all the data from a boltztrap run
    """

    def __init__(self, gap=None, mu_steps=None, cond=None, seebeck=None, kappa=None, hall=None, doping=None,
                 mu_doping=None, seebeck_doping=None, cond_doping=None, kappa_doping=None,
                 hall_doping=None, dos=None, dos_partial=None, carrier_conc=None, vol=None, warning=None,
                 bz_bands=None, bz_kpoints=None, fermi_surface_data=None):
        """
        Constructor taking directly all the data generated by Boltztrap. You
        won't probably use it directly but instead use the from_files and
        from_dict methods.

        Args:
            gap: The gap after interpolation in eV
            mu_steps: The steps of electron chemical potential (or Fermi
                level) in eV.
            cond: The electronic conductivity tensor divided by a constant
                relaxation time (sigma/tau) at different temperature and
                fermi levels.
                The format is {temperature: [array of 3x3 tensors at each
                fermi level in mu_steps]}. The units are 1/(Ohm*m*s).
            seebeck: The Seebeck tensor at different temperatures and fermi
                levels. The format is {temperature: [array of 3x3 tensors at
                each fermi level in mu_steps]}. The units are V/K
            kappa: The electronic thermal conductivity tensor divided by a
                constant relaxation time (kappa/tau) at different temperature
                and fermi levels. The format is {temperature: [array of 3x3
                tensors at each fermi level in mu_steps]}
                The units are W/(m*K*s)
            hall: The hall tensor at different temperature and fermi levels
                The format is {temperature: [array of 27 coefficients list at
                each fermi level in mu_steps]}
                The units are m^3/C
            doping: The different doping levels that have been given to
                Boltztrap. The format is {'p':[],'n':[]} with an array of
                doping levels. The units are cm^-3
            mu_doping: Gives the electron chemical potential (or Fermi level)
                for a given set of doping.
                Format is {'p':{temperature: [fermi levels],'n':{temperature:
                [fermi levels]}}
                the fermi level array is ordered according to the doping
                levels in doping units for doping are in cm^-3 and for Fermi
                level in eV
            seebeck_doping: The Seebeck tensor at different temperatures and
                doping levels. The format is {'p': {temperature: [Seebeck
                tensors]}, 'n':{temperature: [Seebeck tensors]}}
                The [Seebeck tensors] array is ordered according to the
                doping levels in doping units for doping are in cm^-3 and for
                Seebeck in V/K
            cond_doping: The electronic conductivity tensor divided by a
                constant relaxation time (sigma/tau) at different
                temperatures and doping levels
                The format is {'p':{temperature: [conductivity tensors]},
                'n':{temperature: [conductivity tensors]}}
                The [conductivity tensors] array is ordered according to the
                doping levels in doping units for doping are in cm^-3 and for
                conductivity in 1/(Ohm*m*s)
            kappa_doping: The thermal conductivity tensor divided by a constant
                relaxation time (kappa/tau) at different temperatures and
                doping levels.
                The format is {'p':{temperature: [thermal conductivity
                tensors]},'n':{temperature: [thermal conductivity tensors]}}
                The [thermal conductivity tensors] array is ordered according
                to the doping levels in doping units for doping are in cm^-3
                and for thermal conductivity in W/(m*K*s)
            hall_doping: The Hall tensor at different temperatures and doping
                levels.
                The format is {'p':{temperature: [Hall tensors]},
                'n':{temperature: [Hall tensors]}}
                The [Hall tensors] array is ordered according to the doping
                levels in doping and each Hall tensor is represented by a 27
                coefficients list.
                The units are m^3/C
            carrier_conc: The concentration of carriers in electron (or hole)
                per unit cell
            dos: The dos computed by Boltztrap given as a pymatgen Dos object
            dos_partial: Data for the partial DOS projected on sites and
                orbitals
            vol: Volume of the unit cell in angstrom cube (A^3)
            warning: True if Boltztrap spitted out a warning
            bz_bands: Data for interpolated bands on a k-point line (run_type=BANDS)
            bz_kpoints: k-point in reciprocal coordinates for interpolated bands (run_type=BANDS)
            fermi_surface_data: energy values in a 3D grid imported from the output .cube file using ase.io.cube.read_cube
        """
        self.gap = gap
        self.mu_steps = mu_steps
        self._cond = cond
        self._seebeck = seebeck
        self._kappa = kappa
        self._hall = hall
        self.warning = warning
        self.doping = doping
        self.mu_doping = mu_doping
        self._seebeck_doping = seebeck_doping
        self._cond_doping = cond_doping
        self._kappa_doping = kappa_doping
        self._hall_doping = hall_doping
        self._carrier_conc = carrier_conc
        self.dos = dos
        self.vol = vol
        self._dos_partial = dos_partial
        self._bz_bands = bz_bands
        self._bz_kpoints = bz_kpoints
        self.fermi_surface_data = fermi_surface_data

    @staticmethod
    def _make_boltztrap_analyzer_from_data(
            data_full=None, data_hall=None, data_dos=None, temperature_steps=None, mu_steps=None,
            efermi=None, gap=None, doping=None, data_doping_full=None, data_doping_hall=None, vol=None,
            warning=False, bz_bands=None, bz_kpoints=None, run_type="BOLTZ", dos_spin=None, fermi_surface_data=None):
        """
        Make a BoltztrapAnalyzer object from raw data typically parse from
        files.
        """
        if run_type == "BOLTZ":
            cond = {t: [] for t in temperature_steps}
            seebeck = {t: [] for t in temperature_steps}
            kappa = {t: [] for t in temperature_steps}
            hall = {t: [] for t in temperature_steps}
            carrier_conc = {t: [] for t in temperature_steps}
            dos_full = {'energy': [], 'density': []}
            warning = warning
            new_doping = {'p': [], 'n': []}
            for d in doping:
                if d > 0:
                    new_doping['p'].append(d)
                else:
                    new_doping['n'].append(-d)

            mu_doping = {'p': {t: [] for t in temperature_steps},
                         'n': {t: [] for t in temperature_steps}}
            seebeck_doping = {'p': {t: [] for t in temperature_steps},
                              'n': {t: [] for t in temperature_steps}}
            cond_doping = {'p': {t: [] for t in temperature_steps},
                           'n': {t: [] for t in temperature_steps}}
            kappa_doping = {'p': {t: [] for t in temperature_steps},
                            'n': {t: [] for t in temperature_steps}}
            hall_doping = {'p': {t: [] for t in temperature_steps},
                           'n': {t: [] for t in temperature_steps}}
            for d in data_full:
                carrier_conc[d[1]].append(d[2])
                tens_cond = [[d[3], d[4], d[5]],
                             [d[6], d[7], d[8]],
                             [d[9], d[10], d[11]]]
                cond[d[1]].append(tens_cond)
                tens_seebeck = [[d[12], d[13], d[14]],
                                [d[15], d[16], d[17]],
                                [d[18], d[19], d[20]]]
                seebeck[d[1]].append(tens_seebeck)
                tens_kappa = [[d[21], d[22], d[23]],
                              [d[24], d[25], d[26]],
                              [d[27], d[28], d[29]]]
                kappa[d[1]].append(tens_kappa)

            for d in data_hall:
                hall_tens = [[[d[3], d[4], d[5]],
                              [d[6], d[7], d[8]],
                              [d[9], d[10], d[11]]],
                             [[d[12], d[13], d[14]],
                              [d[15], d[16], d[17]],
                              [d[18], d[19], d[20]]],
                             [[d[21], d[22], d[23]],
                              [d[24], d[25], d[26]],
                              [d[27], d[28], d[29]]]]
                hall[d[1]].append(hall_tens)

            for d in data_doping_full:
                tens_cond = [[d[2], d[3], d[4]],
                             [d[5], d[6], d[7]],
                             [d[8], d[9], d[10]]]
                tens_seebeck = [[d[11], d[12], d[13]],
                                [d[14], d[15], d[16]],
                                [d[17], d[18], d[19]]]
                tens_kappa = [[d[20], d[21], d[22]],
                              [d[23], d[24], d[25]],
                              [d[26], d[27], d[28]]]

                if d[1] < 0:
                    mu_doping['n'][d[0]].append(Energy(d[-1], "Ry").to("eV"))
                    cond_doping['n'][d[0]].append(tens_cond)
                    seebeck_doping['n'][d[0]].append(tens_seebeck)
                    kappa_doping['n'][d[0]].append(tens_kappa)
                else:
                    mu_doping['p'][d[0]].append(Energy(d[-1], "Ry").to("eV"))
                    cond_doping['p'][d[0]].append(tens_cond)
                    seebeck_doping['p'][d[0]].append(tens_seebeck)
                    kappa_doping['p'][d[0]].append(tens_kappa)

            for d in data_doping_hall:
                hall_tens = [[[d[2], d[3], d[4]],
                              [d[5], d[6], d[7]],
                              [d[8], d[9], d[10]]],
                             [[d[11], d[12], d[13]],
                              [d[14], d[15], d[16]],
                              [d[17], d[18], d[19]]],
                             [[d[20], d[21], d[22]],
                              [d[23], d[24], d[25]],
                              [d[26], d[27], d[28]]]]
                if d[1] < 0:
                    hall_doping['n'][d[0]].append(hall_tens)
                else:
                    hall_doping['p'][d[0]].append(hall_tens)

            for t in data_dos['total']:
                dos_full['energy'].append(t[0])
                dos_full['density'].append(t[1])

            dos = Dos(efermi, dos_full['energy'], {Spin.up: dos_full['density']})
            dos_partial = data_dos['partial']

            return BoltztrapAnalyzer(
                gap, mu_steps, cond, seebeck, kappa, hall, new_doping, mu_doping,
                seebeck_doping, cond_doping, kappa_doping, hall_doping, dos,
                dos_partial, carrier_conc, vol, warning)

        elif run_type == "BANDS":

            return BoltztrapAnalyzer(bz_bands=bz_bands, bz_kpoints=bz_kpoints)

        elif run_type == "DOS":
            if not dos_spin:
                print(
                    "Spin component set to up (dos_spin=1). Set dos_spin=-1 in from_files function if you want spin down")
                dos_spin = 1

            dos_full = {'energy': [], 'density': []}

            for t in data_dos['total']:
                dos_full['energy'].append(t[0])
                dos_full['density'].append(t[1])

            dos = Dos(efermi, dos_full['energy'], {Spin(dos_spin): dos_full['density']})
            dos_partial = data_dos['partial']

            return BoltztrapAnalyzer(dos=dos, dos_partial=dos_partial)

        elif run_type == "FERMI":
            return BoltztrapAnalyzer(fermi_surface_data=fermi_surface_data)

    def get_symm_bands(self, structure, efermi, kpt_line=None, labels_dict=None):
        """
            Function usefull to read bands from Boltztap output and get a BandStructureSymmLine object comparable with
            that one from a DFT calculation (if the same kpt_line is provided).
            Default kpt_line and labels_dict is the standard path of high symmetry k-point for the specified structure.
            They could be extracted from the BandStructureSymmLine object that you want to compare with.
            efermi variable must be specified to create the BandStructureSymmLine object (usually it comes from DFT or Boltztap calc)
        """
        try:
            if kpt_line is None:
                kpath = HighSymmKpath(structure)
                kpt_line = [Kpoint(k, structure.reciprocal_lattice) for k in
                            kpath.get_kpoints(coords_are_cartesian=False)[0]]
                labels_dict = {l: k for k, l in zip(*kpath.get_kpoints(coords_are_cartesian=False)) if l}
                kpoints = [kp.frac_coords for kp in kpt_line]
            else:
                kpoints = [kp.frac_coords for kp in kpt_line]
                labels_dict = {k: labels_dict[k].frac_coords for k in labels_dict}

            idx_list = []
            #       kpt_dense=np.array([kp for kp in self._bz_kpoints])
            for i, kp in enumerate(kpt_line):
                w = []
                prec = 1e-05
                while len(w) == 0:
                    w = np.where(np.all(np.abs(kp.frac_coords - self._bz_kpoints) < [prec] * 3, axis=1))[0]
                    prec *= 10

                # print( prec )
                idx_list.append([i, w[0]])

                # if len(w)>0:
                #     idx_list.append([i,w[0]])
                # else:
                #     w=np.where(np.all(np.abs(kp.frac_coords-self._bz_kpoints)<[1e-04,1e-04,1e-04],axis=1))[0]
                #     idx_list.append([i,w[0]])

            idx_list = np.array(idx_list)
            # print( idx_list.shape )

            bands_dict = {Spin.up: (self._bz_bands * Energy(1, "Ry").to("eV") + efermi).T[:, idx_list[:, 1]].tolist()}
            # bz_kpoints = bz_kpoints[idx_list[:,1]].tolist()

            sbs = BandStructureSymmLine(kpoints, bands_dict, structure.reciprocal_lattice, efermi,
                                        labels_dict=labels_dict)

            return sbs

        except:
            raise BoltztrapError("Bands are not in output of BoltzTraP.\nBolztrapRunner have to be run with run_type=BANDS")

    def check_acc_bzt_bands(self, sbs_bz, sbs_ref, warn_thr=0.01):
        """
            Compare sbs_bz BandStructureSymmLine calculated with boltztrap with
            the sbs_ref BandStructureSymmLine as reference (from MP for istance),
            using correlation.
            See compare_sym_bands function doc
            Return a list of correlation values of the eigth bands with index
            that ranges from vbm_idx-3 to cbm_idx +3
            Return also two list of sum of relative errors for each branch of vbm and cbm
            and a boolean variable to signal the presence of not accurate bands around the gap.
            warn_thr is a threshold to get a warning in the accuracy of Boltztap interpolated bands.
        """
        if not sbs_ref.is_metal():
            vbm_idx = sbs_ref.get_vbm()['band_index'][Spin.up][-1]
            cbm_idx = sbs_ref.get_cbm()['band_index'][Spin.up][0]
            corr, werr_vbm = compare_sym_bands(sbs_bz, sbs_ref, vbm_idx)
            corr, werr_cbm = compare_sym_bands(sbs_bz, sbs_ref, cbm_idx)

            acc_err = False
            warn_thr = 0.01
            if any(corr[vbm_idx - 3:vbm_idx + 1] > warn_thr) or any(corr[cbm_idx:cbm_idx + 4] > warn_thr):
                print("Warning! some bands around gap are not accurate")
                acc_err = True

            return corr, werr_vbm, werr_cbm, acc_err
        
        else:
            raise BoltztrapError("band check implemented only for bandstructure with gap")

    def get_seebeck(self, output='eig', doping_levels=True):
        """
            Gives the seebeck coefficient in either a full 3x3 tensor form,
            as 3 eigenvalues, or as the average value
            (trace/3.0) If doping_levels=True, the results are given at
            different p and n doping
            levels (given by self.doping), otherwise it is given as a series
            of electron chemical potential values

            Args:
                output (string): the type of output. 'tensor' give the full
                3x3 tensor, 'eig' its 3 eigenvalues and
                'average' the average of the three eigenvalues
                doping_levels (boolean): True for the results to be given at
                different doping levels, False for results
                at different electron chemical potentials

            Returns:
                If doping_levels=True, a dictionnary {temp:{'p':[],'n':[]}}.
                The 'p' links to Seebeck at p-type doping
                and 'n' to the Seebeck at n-type doping. Otherwise, returns a
                {temp:[]} dictionary
                The result contains either the sorted three eigenvalues of
                the symmetric
                Seebeck tensor (output='eig') or a full tensor (3x3 array) (
                output='tensor') or as an average
                (output='average').

                units are microV/K
        """
        return BoltztrapAnalyzer._format_to_output(self._seebeck,
                                                   self._seebeck_doping, output,
                                                   doping_levels, 1e6)

    def get_conductivity(self, output='eig', doping_levels=True,
                         relaxation_time=1e-14):
        """
            Gives the conductivity in either a full 3x3 tensor form,
            as 3 eigenvalues, or as the average value
            (trace/3.0) If doping_levels=True, the results are given at
            different p and n doping
            levels (given by self.doping), otherwise it is given as a series
            of electron chemical potential values

            Args:
                output (string): the type of output. 'tensor' give the full
                3x3 tensor, 'eig' its 3 eigenvalues and
                'average' the average of the three eigenvalues
                doping_levels (boolean): True for the results to be given at
                different doping levels, False for results
                at different electron chemical potentials

            Returns:
                If doping_levels=True, a dictionnary {temp:{'p':[],'n':[]}}.
                The 'p' links to conductivity
                at p-type doping and 'n' to the conductivity at n-type
                doping. Otherwise,
                returns a {temp:[]} dictionary. The result contains either
                the sorted three eigenvalues of the symmetric
                conductivity tensor (format='eig') or a full tensor (3x3
                array) (output='tensor') or as an average
                (output='average').
                The result includes a given constant relaxation time

                units are 1/Ohm*m
        """
        return BoltztrapAnalyzer._format_to_output(self._cond,
                                                   self._cond_doping, output,
                                                   doping_levels,
                                                   relaxation_time)

    def get_power_factor(self, output='eig', doping_levels=True,
                         relaxation_time=1e-14):
        """
        Gives the power factor (Seebeck^2 * conductivity) in either a full
        3x3 tensor form,
        as 3 eigenvalues, or as the average value (trace/3.0) If
        doping_levels=True, the results are given at
        different p and n doping levels (given by self.doping), otherwise it
        is given as a series of
        electron chemical potential values

        Args:
            output (string): the type of output. 'tensor' give the full 3x3
            tensor, 'eig' its 3 eigenvalues and
            'average' the average of the three eigenvalues
            doping_levels (boolean): True for the results to be given at
            different doping levels, False for results
            at different electron chemical potentials

        Returns:
            If doping_levels=True, a dictionnary {temp:{'p':[],'n':[]}}. The
            'p' links to power factor
            at p-type doping and 'n' to the conductivity at n-type doping.
            Otherwise,
            returns a {temp:[]} dictionary. The result contains either the
            sorted three eigenvalues of the symmetric
            power factor tensor (format='eig') or a full tensor (3x3 array) (
            output='tensor') or as an average
            (output='average').
            The result includes a given constant relaxation time

            units are microW/(m K^2)
        """
        result_doping = {doping: {t: [] for t in self._seebeck_doping[doping]}
                         for doping in self._seebeck_doping}
        for doping in result_doping:
            for temp in result_doping[doping]:
                for i in range(len(self.doping[doping])):
                    full_tensor = np.dot(self._cond_doping[doping][temp][i],
                                         np.dot(
                                             self._seebeck_doping[doping][temp][
                                                 i],
                                             self._seebeck_doping[doping][temp][
                                                 i]))
                    result_doping[doping][temp].append(full_tensor)

        result = {temp: [] for temp in self._seebeck}
        for temp in result:
            for i in range(len(self.mu_steps)):
                full_tensor = np.dot(self._cond[temp][i],
                                     np.dot(self._seebeck[temp][i],
                                            self._seebeck[temp][i]))
                result[temp].append(full_tensor)
        return BoltztrapAnalyzer._format_to_output(result, result_doping,
                                                   output, doping_levels,
                                                   multi=1e6 * relaxation_time)

    def get_thermal_conductivity(self, output='eig', doping_levels=True,
                                 relaxation_time=1e-14):
        """
        Gives the electronic part of the thermal conductivity in either a
        full 3x3 tensor form,
        as 3 eigenvalues, or as the average value (trace/3.0) If
        doping_levels=True, the results are given at
        different p and n doping levels (given by self.doping), otherwise it
        is given as a series of
        electron chemical potential values

        Args:
            output (string): the type of output. 'tensor' give the full 3x3
            tensor, 'eig' its 3 eigenvalues and
            'average' the average of the three eigenvalues
            doping_levels (boolean): True for the results to be given at
            different doping levels, False for results
            at different electron chemical potentials

        Returns:
            If doping_levels=True, a dictionary {temp:{'p':[],'n':[]}}. The
            'p' links to thermal conductivity
            at p-type doping and 'n' to the thermal conductivity at n-type
            doping. Otherwise,
            returns a {temp:[]} dictionary. The result contains either the
            sorted three eigenvalues of the symmetric
            conductivity tensor (format='eig') or a full tensor (3x3 array) (
            output='tensor') or as an average
            (output='average').
            The result includes a given constant relaxation time

            units are W/mK
        """
        result_doping = {doping: {t: [] for t in self._seebeck_doping[doping]}
                         for doping in self._seebeck_doping}
        for doping in result_doping:
            for temp in result_doping[doping]:
                for i in range(len(self.doping[doping])):
                    pf_tensor = np.dot(self._cond_doping[doping][temp][i],
                                       np.dot(
                                           self._seebeck_doping[doping][temp][
                                               i],
                                           self._seebeck_doping[doping][temp][
                                               i]))
                    result_doping[doping][temp].append((self._kappa_doping[
                                                            doping][temp][
                                                            i] - pf_tensor *
                                                        temp))

        result = {temp: [] for temp in self._seebeck}
        for temp in result:
            for i in range(len(self.mu_steps)):
                pf_tensor = np.dot(self._cond[temp][i],
                                   np.dot(self._seebeck[temp][i],
                                          self._seebeck[temp][i]))
                result[temp].append((self._kappa[temp][i] - pf_tensor * temp))

        return BoltztrapAnalyzer._format_to_output(result, result_doping,
                                                   output, doping_levels,
                                                   multi=relaxation_time)

    def get_zt(self, output='eig', doping_levels=True, relaxation_time=1e-14,
               kl=0.2):
        """
        Gives the ZT coefficient (S^2*cond*T/thermal cond) in either a full
        3x3 tensor form,
        as 3 eigenvalues, or as the average value (trace/3.0) If
        doping_levels=True, the results are given at
        different p and n doping levels (given by self.doping), otherwise it
        is given as a series of
        electron chemical potential values. We assume a constant relaxation
        time and a constant
        lattice thermal conductivity

        Args:
            output (string): the type of output. 'tensor' give the full 3x3
            tensor, 'eig' its 3 eigenvalues and
            'average' the average of the three eigenvalues
            doping_levels (boolean): True for the results to be given at
            different doping levels, False for results
            at different electron chemical potentials

        Returns:
            If doping_levels=True, a dictionary {temp:{'p':[],'n':[]}}. The
            'p' links to ZT
            at p-type doping and 'n' to the ZT at n-type doping. Otherwise,
            returns a {temp:[]} dictionary. The result contains either the
            sorted three eigenvalues of the symmetric
            ZT tensor (format='eig') or a full tensor (3x3 array) (
            output='tensor') or as an average
            (output='average').
            The result includes a given constant relaxation time and lattice
            thermal conductivity
        """
        result_doping = {doping: {t: [] for t in self._seebeck_doping[doping]}
                         for doping in self._seebeck_doping}
        for doping in result_doping:
            for temp in result_doping[doping]:
                for i in range(len(self.doping[doping])):
                    pf_tensor = np.dot(self._cond_doping[doping][temp][i],
                                       np.dot(
                                           self._seebeck_doping[doping][temp][
                                               i],
                                           self._seebeck_doping[doping][temp][
                                               i]))
                    thermal_conduct = (self._kappa_doping[doping][temp][
                                           i] - pf_tensor * temp) * relaxation_time
                    result_doping[doping][temp].append(
                        np.dot(pf_tensor * relaxation_time * temp,
                               np.linalg.inv(
                                   thermal_conduct + kl * np.eye(3, 3))))

        result = {temp: [] for temp in self._seebeck}
        for temp in result:
            for i in range(len(self.mu_steps)):
                pf_tensor = np.dot(self._cond[temp][i],
                                   np.dot(self._seebeck[temp][i],
                                          self._seebeck[temp][i]))
                thermal_conduct = (self._kappa[temp][
                                       i] - pf_tensor * temp) * relaxation_time
                result[temp].append(np.dot(pf_tensor * relaxation_time * temp,
                                           np.linalg.inv(
                                               thermal_conduct + kl * np.eye(3,
                                                                             3))))
        return BoltztrapAnalyzer._format_to_output(result, result_doping,
                                                   output, doping_levels)

    def get_average_eff_mass(self, output='eig'):
        """
        Gives the average effective mass tensor. We call it average because
        it takes into account all the bands
        and regons in the Brillouin zone. This is different than the standard
        textbook effective mass which relates
        often to only one (parabolic) band.
        The average effective mass tensor is defined as the integrated
        average of the second derivative of E(k)
        This effective mass tensor takes into account:
        -non-parabolicity
        -multiple extrema
        -multiple bands

        For more information about it. See:

        Hautier, G., Miglio, A., Waroquiers, D., Rignanese, G., & Gonze,
        X. (2014).
        How Does Chemistry Influence Electron Effective Mass in Oxides?
        A High-Throughput Computational Analysis. Chemistry of Materials,
        26(19), 5447–5458. doi:10.1021/cm404079a

        or

        Hautier, G., Miglio, A., Ceder, G., Rignanese, G.-M., & Gonze,
        X. (2013).
        Identification and design principles of low hole effective mass
        p-type transparent conducting oxides.
        Nature Communications, 4, 2292. doi:10.1038/ncomms3292

        Depending on the value of output, we have either the full 3x3
        effective mass tensor,
        its 3 eigenvalues or an average

        Args:
            output (string): 'eigs' for eigenvalues, 'tensor' for the full
            tensor and 'average' for an average (trace/3)

        Returns:
            a dictionary {'p':{temp:[]},'n':{temp:[]}} with an array of
            effective mass tensor, eigenvalues of average
            value (depending on output) for each temperature and for each
            doping level.
            The 'p' links to hole effective mass tensor and 'n' to electron
            effective mass tensor.
        """

        result_doping = {doping: {t: [] for t in self._cond_doping[doping]} for
                         doping in self.doping}
        for doping in result_doping:
            for temp in result_doping[doping]:
                for i in range(len(self.doping[doping])):
                    if output == 'tensor':
                        result_doping[doping][temp].append(np.linalg.inv(
                            np.array(self._cond_doping[doping][temp][i])) \
                                                           *
                                                           self.doping[doping][
                                                               i] * 10 ** 6 *
                                                           e ** 2 /
                                                           m_e)
                    elif output == 'eig':
                        result_doping[doping][temp].append(
                            sorted(np.linalg.eigh(np.linalg.inv(
                                np.array(self._cond_doping[doping][temp][i])) *
                                                  self.doping[doping][
                                                      i] * 10 ** 6 * e ** 2 \
                                                  / m_e)[0]))
                    else:
                        full_tensor = np.linalg.inv(
                            np.array(self._cond_doping[doping][temp][i])) \
                                      * self.doping[doping][
                                          i] * 10 ** 6 * e ** 2 / m_e
                        result_doping[doping][temp].append((full_tensor[0][0] \
                                                            + full_tensor[1][
                                                                1] \
                                                            + full_tensor[2][
                                                                2]) / 3.0)
        return result_doping

    @staticmethod
    def _format_to_output(tensor, tensor_doping, output, doping_levels,
                          multi=1.0):
        if doping_levels:
            full_tensor = tensor_doping
            result = {doping: {t: [] for t in tensor_doping[doping]} for doping
                      in tensor_doping}
            for doping in full_tensor:
                for temp in full_tensor[doping]:
                    for i in range(len(full_tensor[doping][temp])):
                        if output == 'eig':
                            result[doping][temp].append(sorted(
                                np.linalg.eigh(full_tensor[doping][temp][i])[
                                    0] * multi))
                        elif output == 'tensor':
                            result[doping][temp].append(
                                np.array(full_tensor[doping][temp][i]) * multi)
                        else:
                            result[doping][temp].append(
                                (full_tensor[doping][temp][i][0][0] \
                                 + full_tensor[doping][temp][i][1][1] \
                                 + full_tensor[doping][temp][i][2][
                                     2]) * multi / 3.0)
        else:
            full_tensor = tensor
            result = {t: [] for t in tensor}
            for temp in full_tensor:
                for i in range(len(tensor[temp])):
                    if output == 'eig':
                        result[temp].append(sorted(
                            np.linalg.eigh(full_tensor[temp][i])[0] * multi))
                    elif output == 'tensor':
                        result[temp].append(
                            np.array(full_tensor[temp][i]) * multi)
                    else:
                        result[temp].append((full_tensor[temp][i][0][0] \
                                             + full_tensor[temp][i][1][1] \
                                             + full_tensor[temp][i][2][
                                                 2]) * multi / 3.0)
        return result

    def get_complete_dos(self, structure, analyzer_for_second_spin=None):
        """
            Gives a CompleteDos object with the DOS from the interpolated projected band structure
            Args:
                the structure (necessary to identify sites for projection)
                analyzer_for_second_spin must be specified to have a CompleteDos with both Spin components
            Returns:
                a CompleteDos object
            Example of use in case of spin polarized case:
            
                BoltztrapRunner(bs=bs,nelec=10,run_type="DOS",spin=1).run(path_dir='dos_up/')
                an_up=BoltztrapAnalyzer.from_files("dos_up/boltztrap/",dos_spin=1)

                BoltztrapRunner(bs=bs,nelec=10,run_type="DOS",spin=-1).run(path_dir='dos_dw/')
                an_dw=BoltztrapAnalyzer.from_files("dos_dw/boltztrap/",dos_spin=-1)

                cdos=an_up.get_complete_dos(bs._structure,an_dw)

        """
        pdoss = {}
        spin_1 = self.dos.densities.keys()[0]

        if analyzer_for_second_spin:
            if not np.all(self.dos.energies == analyzer_for_second_spin.dos.energies):
                raise BoltztrapError("Dos merging error: energies of the two dos are different")

            spin_2 = analyzer_for_second_spin.dos.densities.keys()[0]
            if spin_1 == spin_2:
                raise BoltztrapError("Dos merging error: spin component are the same")

        for s in self._dos_partial:
            if structure.sites[int(s)] not in pdoss:
                pdoss[structure.sites[int(s)]] = {}
            for o in self._dos_partial[s]:
                if Orbital[o] not in pdoss[structure.sites[int(s)]]:
                    pdoss[structure.sites[int(s)]][Orbital[o]] = {}
                pdoss[structure.sites[int(s)]][Orbital[o]][
                    spin_1] = self._dos_partial[s][o]
                if analyzer_for_second_spin:
                    pdoss[structure.sites[int(s)]][Orbital[o]][
                        spin_2] = analyzer_for_second_spin._dos_partial[s][o]
        if analyzer_for_second_spin:
            tdos = Dos(self.dos.efermi, self.dos.energies,
                       {spin_1: self.dos.densities[spin_1], spin_2: analyzer_for_second_spin.dos.densities[spin_2]})
        else:
            tdos = self.dos

        return CompleteDos(structure, total_dos=tdos, pdoss=pdoss)

    def get_mu_bounds(self, temp=300):
        return min(self.mu_doping['p'][temp]), max(self.mu_doping['n'][temp])

    def get_carrier_concentration(self):
        """
        gives the carrier concentration (in cm^-3)

        Returns
            a dictionary {temp:[]} with an array of carrier concentration (in cm^-3) at each temperature
            The array relates to each step of electron chemical potential
        """

        return {temp: [1e24 * i / self.vol for i in self._carrier_conc[temp]] for temp in self._carrier_conc}

    def get_hall_carrier_concentration(self):
        """
        gives the Hall carrier concentration (in cm^-3). This is the trace of the Hall tensor (see Boltztrap source
        code) Hall carrier concentration are not always exactly the same than carrier concentration.

        Returns
            a dictionary {temp:[]} with an array of Hall carrier concentration (in cm^-3) at each temperature
            The array relates to each step of electron chemical potential
        """
        result = {temp: [] for temp in self._hall}
        for temp in self._hall:
            for i in self._hall[temp]:
                trace = (i[1][2][0] + i[2][0][1] + i[0][1][2]) / 3.0
                if trace != 0.0:
                    result[temp].append(1e-6 / (trace * e))
                else:
                    result[temp].append(0.0)
        return result

    @staticmethod
    def from_files(path_dir, dos_spin=None):
        """
        get a BoltztrapAnalyzer object from a set of files

        Args:
            path_dir: directory where the boltztrap files are

        Returns:
            a BoltztrapAnalyzer object

        """

        with open(os.path.join(path_dir, "boltztrap.outputtrans"), 'r') as f:
            for line in f:
                if "Calc type:" in line:
                    run_type = line.split()[-1]
                    print(run_type, " calc type found")
                    break

        if run_type == "BOLTZ":
            t_steps = set()
            m_steps = set()
            gap = None
            doping = []
            data_doping_full = []
            data_doping_hall = []
            with open(os.path.join(path_dir, "boltztrap.condtens"), 'r') as f:
                data_full = []
                for line in f:
                    if not line.startswith("#"):
                        t_steps.add(int(float(line.split()[1])))
                        m_steps.add(float(line.split()[0]))
                        data_full.append([float(c) for c in line.split()])

            with open(os.path.join(path_dir, "boltztrap.halltens"), 'r') as f:
                data_hall = []
                for line in f:
                    if not line.startswith("#"):
                        data_hall.append([float(c) for c in line.split()])

            data_dos = {'total': [], 'partial': {}}
            with open(os.path.join(path_dir, "boltztrap.transdos"), 'r') as f:
                count_series = 0
                for line in f:
                    if not line.startswith(" #"):
                        data_dos['total'].append(
                            [Energy(float(line.split()[0]), "Ry").to("eV"),
                             float(line.split()[1])])
                        total_elec = float(line.split()[2])
                    else:
                        count_series += 1
                    if count_series > 1:
                        break
            data_dos['total'] = [
                [data_dos['total'][i][0], 2 * data_dos['total'][i][1] / total_elec]
                for i in range(len(data_dos['total']))]
            for file_name in os.listdir(path_dir):
                if file_name.endswith(
                        "transdos") and file_name != 'boltztrap.transdos':
                    tokens = file_name.split(".")[1].split("_")
                    with open(os.path.join(path_dir, file_name), 'r') as f:
                        for line in f:
                            if not line.startswith(" #"):
                                if tokens[1] not in data_dos['partial']:
                                    data_dos['partial'][tokens[1]] = {}
                                if tokens[2] not in data_dos['partial'][tokens[1]]:
                                    data_dos['partial'][tokens[1]][tokens[2]] = []
                                data_dos['partial'][tokens[1]][tokens[2]].append(
                                    2 * float(line.split()[1]))

            with open(os.path.join(path_dir, "boltztrap.outputtrans"), 'r') as f:
                warning = False
                step = 0
                for line in f:
                    if "WARNING" in line:
                        warning = True
                    if line.startswith("VBM"):
                        efermi = Energy(line.split()[1], "Ry").to("eV")
                    if line.startswith("Doping level number"):
                        doping.append(float(line.split()[6]))
                    if line.startswith("Egap:"):
                        gap = float(line.split()[1])
            if len(doping) != 0:
                with open(os.path.join(path_dir, "boltztrap.condtens_fixdoping"),
                          'r') as f:
                    for line in f:
                        if not line.startswith("#") and len(line) > 2:
                            data_doping_full.append([float(c)
                                                     for c in line.split()])

                with open(os.path.join(path_dir, "boltztrap.halltens_fixdoping"),
                          'r') as f:
                    for line in f:
                        if not line.startswith("#") and len(line) > 2:
                            data_doping_hall.append(
                                [float(c) for c in line.split()])

            with open(os.path.join(path_dir, "boltztrap.struct"), 'r') as f:
                tokens = f.readlines()
                vol = Lattice([[Length(float(tokens[i].split()[j]), "bohr").to("ang")
                                for j in range(3)] for i in range(1, 4)]).volume

            data_dos = {'total': [], 'partial': {}}
            with open(os.path.join(path_dir, "boltztrap.transdos"), 'r') as f:
                count_series = 0
                normalization_factor = None
                for line in f:
                    if not line.startswith(" #"):
                        data_dos['total'].append([Energy(float(line.split()[0]), "Ry").to("eV"),
                                                  float(line.split()[1])])
                    else:
                        count_series += 1
                    if count_series > 1:
                        break
            for file_name in os.listdir(path_dir):
                if file_name.endswith("transdos") and file_name != 'boltztrap.transdos':
                    tokens = file_name.split(".")[1].split("_")
                    with open(os.path.join(path_dir, file_name), 'r') as f:
                        for line in f:
                            if not line.startswith(" #"):
                                if tokens[1] not in data_dos['partial']:
                                    data_dos['partial'][tokens[1]] = {}
                                if tokens[2] not in data_dos['partial'][tokens[1]]:
                                    data_dos['partial'][tokens[1]][tokens[2]] = []
                                data_dos['partial'][tokens[1]][tokens[2]].append(float(line.split()[1]))

            return BoltztrapAnalyzer._make_boltztrap_analyzer_from_data(
                data_full, data_hall, data_dos, sorted([t for t in t_steps]),
                sorted([Energy(m, "Ry").to("eV") for m in m_steps]), efermi,
                Energy(gap, "Ry").to("eV"),
                doping, data_doping_full, data_doping_hall, vol, warning)

        elif run_type == "FERMI":
            from ase.io.cube import read_cube
            if os.path.exists(os.path.join(path_dir, 'boltztrap_BZ.cube')):
                fs_data = read_cube(str(os.path.join(path_dir, 'boltztrap_BZ.cube')), read_data=True)
            if os.path.exists(os.path.join(path_dir, 'fort.30')):
                fs_data = read_cube(str(os.path.join(path_dir, 'fort.30')), read_data=True)
            else:
                raise BoltztrapError("No data file found for fermi surface")
            
            return BoltztrapAnalyzer._make_boltztrap_analyzer_from_data(run_type="FERMI", fermi_surface_data=fs_data)

        elif run_type == "BANDS":
            bz_kpoints = np.loadtxt(os.path.join(path_dir, "boltztrap_band.dat"))[:, -3:]
            bz_bands = np.loadtxt(os.path.join(path_dir, "boltztrap_band.dat"))[:, 1:-6]
            return BoltztrapAnalyzer._make_boltztrap_analyzer_from_data(run_type="BANDS",
                                                                        bz_bands=bz_bands, bz_kpoints=bz_kpoints)

        elif run_type == "DOS":
            data_dos = {'total': [], 'partial': {}}
            with open(os.path.join(path_dir, "boltztrap.transdos"), 'r') as f:
                count_series = 0
                for line in f:
                    if not line.startswith(" #"):
                        data_dos['total'].append([Energy(float(line.split()[0]), "Ry").to("eV"),
                                                  float(line.split()[1])])
                    else:
                        count_series += 1
                    if count_series > 1:
                        break
            print("here")
            tmp_data = np.array(data_dos['total'])
            tmp_den = np.trim_zeros(tmp_data[:,1],'f')[1:]
            lw_l = len(tmp_data[:,1])-len(tmp_den)
            tmp_ene = tmp_data[lw_l:,0]
            tmp_den = np.trim_zeros(tmp_den,'b')[:-1]
            hg_l = len(tmp_ene)-len(tmp_den)
            tmp_ene = tmp_ene[:-hg_l]
            tmp_data = np.vstack((tmp_ene,tmp_den)).T
            data_dos['total'] = tmp_data.tolist()

            for file_name in os.listdir(path_dir):
                if file_name.endswith("transdos") and file_name != 'boltztrap.transdos':
                    tokens = file_name.split(".")[1].split("_")
                    site = tokens[1]
                    orb = '_'.join(tokens[2:])
                    with open(os.path.join(path_dir, file_name), 'r') as f:
                        for line in f:
                            if not line.startswith(" #"):
                                if site not in data_dos['partial']:
                                    data_dos['partial'][site] = {}
                                if orb not in data_dos['partial'][site]:
                                    data_dos['partial'][site][orb] = []
                                data_dos['partial'][site][orb].append(float(line.split()[1]))
                    data_dos['partial'][site][orb] = data_dos['partial'][site][orb][lw_l:-hg_l]
            
            with open(os.path.join(path_dir, "boltztrap.outputtrans"), 'r') as f:
                for line in f:
                    if line.startswith("VBM"):
                        efermi = Energy(line.split()[1], "Ry").to("eV")

            return BoltztrapAnalyzer._make_boltztrap_analyzer_from_data(efermi=efermi, run_type="DOS",
                                                                        data_dos=data_dos, dos_spin=dos_spin)

    def as_dict(self):

        results = {'gap': self.gap,
                   'mu_steps': self.mu_steps,
                   'cond': self._cond,
                   'seebeck': self._seebeck,
                   'kappa': self._kappa,
                   'hall': self._hall,
                   'warning': self.warning, 'doping': self.doping,
                   'mu_doping': self.mu_doping,
                   'seebeck_doping': self._seebeck_doping,
                   'cond_doping': self._cond_doping,
                   'kappa_doping': self._kappa_doping,
                   'hall_doping': self._hall_doping,
                   'dos': self.dos.as_dict(),
                   'dos_partial': self._dos_partial,
                   'carrier_conc': self._carrier_conc,
                   'vol': self.vol}
        return jsanitize(results)

    @staticmethod
    def from_dict(data):
        def _make_float_array(a):
            res = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]
            for i in range(3):
                for j in range(3):
                    res[i][j] = float(a[i][j])
            return res

        def _make_float_hall(a):
            return [i for i in a[:27]]

        return BoltztrapAnalyzer(
            float(data['gap']), [float(d) for d in data['mu_steps']],
            {int(d): [_make_float_array(v) for v in data['cond'][d]]
             for d in data['cond']},
            {int(d): [_make_float_array(v) for v in data['seebeck'][d]]
             for d in data['seebeck']},
            {int(d): [_make_float_array(v) for v in data['kappa'][d]]
             for d in data['kappa']},
            {int(d): [_make_float_hall(v) for v in data['hall'][d]]
             for d in data['hall']},
            {'p': [float(d) for d in data['doping']['p']],
             'n': [float(d) for d in data['doping']['n']]},
            {'p': {int(d): [float(v) for v in data['mu_doping']['p'][d]]
                   for d in data['mu_doping']['p']},
             'n': {int(d): [float(v) for v in data['mu_doping']['n'][d]]
                   for d in data['mu_doping']['n']}},
            {'p': {int(d): [_make_float_array(v)
                            for v in data['seebeck_doping']['p'][d]]
                   for d in data['seebeck_doping']['p']},
             'n': {int(d): [_make_float_array(v)
                            for v in data['seebeck_doping']['n'][d]]
                   for d in data['seebeck_doping']['n']}},
            {'p': {int(d): [_make_float_array(v)
                            for v in data['cond_doping']['p'][d]]
                   for d in data['cond_doping']['p']},
             'n': {int(d): [_make_float_array(v)
                            for v in data['cond_doping']['n'][d]]
                   for d in data['cond_doping']['n']}},
            {'p': {int(d): [_make_float_array(v)
                            for v in data['kappa_doping']['p'][d]]
                   for d in data['kappa_doping']['p']},
             'n': {int(d): [_make_float_array(v)
                            for v in data['kappa_doping']['n'][d]]
                   for d in data['kappa_doping']['n']}},
            {'p': {int(d): [_make_float_hall(v)
                            for v in data['hall_doping']['p'][d]]
                   for d in data['hall_doping']['p']},
             'n': {int(d): [_make_float_hall(v)
                            for v in data['hall_doping']['n'][d]]
                   for d in data['hall_doping']['n']}},
            Dos.from_dict(data['dos']), data['dos_partial'],
            data['carrier_conc'],
            data['vol'], str(data['warning']))


class BoltztrapPlotter(object):
    """
    class containing methods to plot the data from Boltztrap.

    Args:
        bz: a BoltztrapAnalyzer object
    """

    def __init__(self, bz):
        self._bz = bz

    def _plot_doping(self, temp):
        if len(self._bz.doping) != 0:
            limit = 2.21e15
            plt.axvline(self._bz.mu_doping['n'][temp][0], linewidth=3.0,
                        linestyle="--")
            plt.text(self._bz.mu_doping['n'][temp][0] + 0.01,
                     limit,
                     "$n$=10$^{" + str(
                         math.log10(self._bz.doping['n'][0])) + "}$",
                     color='b')
            plt.axvline(self._bz.mu_doping['n'][temp][-1], linewidth=3.0,
                        linestyle="--")
            plt.text(self._bz.mu_doping['n'][temp][-1] + 0.01,
                     limit,
                     "$n$=10$^{" + str(math.log10(self._bz.doping['n'][-1]))
                     + "}$", color='b')
            plt.axvline(self._bz.mu_doping['p'][temp][0], linewidth=3.0,
                        linestyle="--")
            plt.text(self._bz.mu_doping['p'][temp][0] + 0.01,
                     limit,
                     "$p$=10$^{" + str(
                         math.log10(self._bz.doping['p'][0])) + "}$",
                     color='b')
            plt.axvline(self._bz.mu_doping['p'][temp][-1], linewidth=3.0,
                        linestyle="--")
            plt.text(self._bz.mu_doping['p'][temp][-1] + 0.01,
                     limit, "$p$=10$^{" +
                     str(math.log10(self._bz.doping['p'][-1])) + "}$",
                     color='b')

    def _plot_bg_limits(self):
        plt.axvline(0.0, color='k', linewidth=3.0)
        plt.axvline(self._bz.gap, color='k', linewidth=3.0)

    def plot_seebeck_mu(self, temp=600, output='eig', xlim=None):
        """
        Plot the seebeck coefficient in function of Fermi level

        Args:
            temp:
                the temperature
            xlim:
                a list of min and max fermi energy by default (0, and band gap)
        Returns:
            a matplotlib object
        """
        seebeck = self._bz.get_seebeck(output=output, doping_levels=False)[temp]
        plt.plot(self._bz.mu_steps, seebeck,
                 linewidth=3.0)
        self._plot_bg_limits()
        self._plot_doping(temp)
        if output == 'eig':
            plt.legend(['S$_1$', 'S$_2$', 'S$_3$'])
        if xlim is None:
            plt.xlim(-0.5, self._bz.gap + 0.5)
        else:
            plt.xlim(xlim[0], xlim[1])
        plt.ylabel("Seebeck \n coefficient  ($\mu$V/K)", fontsize=30.0)
        plt.xlabel("E-E$_f$ (eV)", fontsize=30)
        plt.xticks(fontsize=25)
        plt.yticks(fontsize=25)
        return plt

    def plot_conductivity_mu(self, temp=600, output='eig',
                             relaxation_time=1e-14, xlim=None):
        """
        Plot the conductivity in function of Fermi level. Semi-log plot

        Args:
            temp: the temperature
            xlim: a list of min and max fermi energy by default (0, and band
                gap)
            tau: A relaxation time in s. By default none and the plot is by
               units of relaxation time

        Returns:
            a matplotlib object
        """
        cond = self._bz.get_conductivity(relaxation_time=relaxation_time,
                                         output=output, doping_levels=False)[
            temp]
        plt.semilogy(self._bz.mu_steps, cond, linewidth=3.0)
        self._plot_bg_limits()
        self._plot_doping(temp)
        if output == 'eig':
            plt.legend(['$\sigma_1$', '$\sigma_2$', '$\sigma_3$'])
        if xlim is None:
            plt.xlim(-0.5, self._bz.gap + 0.5)
        else:
            plt.xlim(xlim)
        plt.ylim([1e13 * relaxation_time, 1e20 * relaxation_time])
        plt.ylabel("conductivity,\n $\sigma$ (1/($\Omega$ m))", fontsize=30.0)
        plt.xlabel("E-E$_f$ (eV)", fontsize=30.0)
        plt.xticks(fontsize=25)
        plt.yticks(fontsize=25)
        return plt

    def plot_power_factor_mu(self, temp=600, output='eig',
                             relaxation_time=1e-14, xlim=None):
        """
        Plot the power factor in function of Fermi level. Semi-log plot

        Args:
            temp: the temperature
            xlim: a list of min and max fermi energy by default (0, and band
                gap)
            tau: A relaxation time in s. By default none and the plot is by
               units of relaxation time

        Returns:
            a matplotlib object
        """
        pf = self._bz.get_power_factor(relaxation_time=relaxation_time,
                                       output=output, doping_levels=False)[temp]
        plt.semilogy(self._bz.mu_steps, pf, linewidth=3.0)
        self._plot_bg_limits()
        self._plot_doping(temp)
        if output == 'eig':
            plt.legend(['PF$_1$', 'PF$_2$', 'PF$_3$'])
        if xlim is None:
            plt.xlim(-0.5, self._bz.gap + 0.5)
        else:
            plt.xlim(xlim)
        plt.ylabel("Power factor, ($\mu$W/(mK$^2$))", fontsize=30.0)
        plt.xlabel("E-E$_f$ (eV)", fontsize=30.0)
        plt.xticks(fontsize=25)
        plt.yticks(fontsize=25)
        return plt

    def plot_zt_mu(self, temp=600, output='eig', relaxation_time=1e-14,
                   xlim=None):
        """
        Plot the ZT in function of Fermi level.

        Args:
            temp: the temperature
            xlim: a list of min and max fermi energy by default (0, and band
                gap)
            tau: A relaxation time in s. By default none and the plot is by
               units of relaxation time

        Returns:
            a matplotlib object
        """
        zt = self._bz.get_zt(relaxation_time=relaxation_time, output=output,
                             doping_levels=False)[temp]
        plt.plot(self._bz.mu_steps, zt, linewidth=3.0)
        self._plot_bg_limits()
        self._plot_doping(temp)
        if output == 'eig':
            plt.legend(['ZT$_1$', 'ZT$_2$', 'ZT$_3$'])
        if xlim is None:
            plt.xlim(-0.5, self._bz.gap + 0.5)
        else:
            plt.xlim(xlim)
        plt.ylabel("ZT", fontsize=30.0)
        plt.xlabel("E-E$_f$ (eV)", fontsize=30.0)
        plt.xticks(fontsize=25)
        plt.yticks(fontsize=25)
        return plt

    def plot_dos(self, sigma=0.05):
        """
        plot dos

        Args:
            sigma: a smearing

        Returns:
            a matplotlib object
        """
        plotter = DosPlotter(sigma=sigma)
        plotter.add_dos("t", self._bz.dos)
        return plotter.get_plot()

    def plot_carriers(self, temp=300):
        """
        Plot the carrier concentration in function of Fermi level

        Args:
            temp: the temperature

        Returns:
            a matplotlib object
        """
        plt.semilogy(self._bz.mu_steps,
                     abs(self._bz.carrier_conc[temp] / (self._bz.vol * 1e-24)),
                     linewidth=3.0, color='r')
        self._plot_bg_limits()
        self._plot_doping(temp)
        plt.xlim(-0.5, self._bz.gap + 0.5)
        plt.ylim(1e14, 1e22)
        plt.ylabel("carrier concentration (cm-3)", fontsize=30.0)
        plt.xlabel("E-E$_f$ (eV)", fontsize=30)
        plt.xticks(fontsize=25)
        plt.yticks(fontsize=25)
        return plt

    def plot_hall_carriers(self, temp=300):
        """
        Plot the Hall carrier concentration in function of Fermi level

        Args:
            temp: the temperature

        Returns:
            a matplotlib object
        """
        hall_carriers = [abs(i) for i in self._bz.get_hall_carrier_concentration()[temp]]
        plt.semilogy(self._bz.mu_steps,
                     hall_carriers,
                     linewidth=3.0, color='r')
        self._plot_bg_limits()
        self._plot_doping(temp)
        plt.xlim(-0.5, self._bz.gap + 0.5)
        plt.ylim(1e14, 1e22)
        plt.ylabel("Hall carrier concentration (cm-3)", fontsize=30.0)
        plt.xlabel("E-E$_f$ (eV)", fontsize=30)
        plt.xticks(fontsize=25)
        plt.yticks(fontsize=25)
        return plt

    def plot_fermi_surface(self, structure=None, isolevel=None):
        """
        Plot the Fermi surface at a aspecific energy value

        Args:
            bz_lattice: structure object of the material
            isolevel: energy value fo fermi surface, Default: max energy value + 0.1eV

        Returns:
            a matplotlib object

        Note: Experimental
        """
        from mpl_toolkits.mplot3d import Axes3D
        from pymatgen.electronic_structure.plotter import plot_brillouin_zone
        try:
            from skimage import measure
        except ImportError:
            raise BoltztrapError("skimage package should be installed to use this function")
        
        fig = None

        data = self._bz.fermi_surface_data

        if not isolevel:
            isolevel = max(data[0].flat) - Energy(0.1, "eV").to("Ry")

        verts, faces = measure.marching_cubes(data[0], isolevel)
        verts -= 1
        verts2 = np.dot(verts, data[1].cell / np.array(data[0].shape)[:, np.newaxis])
        verts2 /= max(verts2.flat) / 1.5

        cx, cy, cz = [(max(verts2[:, i]) - min(verts2[:, i])) / 2 + min(verts2[:, i]) for i in range(3)]

        if structure is not None:
            kpath = HighSymmKpath(structure).kpath
            lines = [[kpath['kpoints'][k] for k in p] for p in kpath['path']]
            fig = plot_brillouin_zone(bz_lattice=structure.reciprocal_lattice, lines=lines, labels=kpath['kpoints'])

        if fig:
            ax = fig.gca()
            ax.plot_trisurf(verts2[:, 0] - cx, verts2[:, 1] - cy, faces, verts2[:, 2] - cz, lw=0)
        else:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
            ax.plot_trisurf(verts2[:, 0] - cx, verts2[:, 1] - cy, faces, verts2[:, 2] - cz, lw=0)
            ax.set_xlim3d(-1, 1)
            ax.set_ylim3d(-1, 1)
            ax.set_zlim3d(-1, 1)
            ax.set_aspect('equal')
            ax.axis("off")

        return fig, ax


def compare_sym_bands(bands_obj, bands_ref_obj, nb=None):
    """
        Compute the mean of correlation between bzt and vasp bandstructure on sym line,
        for all bands and locally (for each branches) the difference squared (%) if nb is specified.
    """

    nkpt = len(bands_obj.kpoints)
    if bands_ref_obj._is_spin_polarized:
        nbands = np.min((bands_obj.nb_bands, 2 * bands_ref_obj.nb_bands))
    else:
        nbands = np.min((len(bands_obj.bands[Spin.up]), len(bands_ref_obj.bands[Spin.up])))
    # print(nbands)
    arr_bands = np.array(bands_obj.bands[Spin.up][:nbands])
    # arr_bands_lavg = (arr_bands-np.mean(arr_bands,axis=1).reshape(nbands,1))

    if bands_ref_obj._is_spin_polarized:
        arr_bands_ref_up = np.array(bands_ref_obj.bands[Spin.up])
        arr_bands_ref_dw = np.array(bands_ref_obj.bands[Spin.down])
        # print(arr_bands_ref_up.shape)
        arr_bands_ref = np.vstack((arr_bands_ref_up, arr_bands_ref_dw))
        arr_bands_ref = np.sort(arr_bands_ref, axis=0)[:nbands]
        # print(arr_bands_ref.shape)
    else:
        arr_bands_ref = np.array(bands_ref_obj.bands[Spin.up][:nbands])

    # arr_bands_ref_lavg = (arr_bands_ref-np.mean(arr_bands_ref,axis=1).reshape(nbands,1))

    # err = np.sum((arr_bands_lavg-arr_bands_ref_lavg)**2,axis=1)/nkpt
    corr = np.array([distance.correlation(arr_bands[idx], arr_bands_ref[idx]) for idx in range(nbands)])

    if type(nb) == int and nb < nbands:
        branches = [[s['start_index'], s['end_index'], s['name']] for s in bands_ref_obj._branches]
        werr = {}
        for start, end, name in branches:
            # werr.append((sum((arr_bands_corr[nb][start:end+1] - arr_bands_ref_corr[nb][start:end+1])**2)/(end+1-start)*100,name))
            werr[name] = np.sum(abs(arr_bands[nb][start:end + 1] - arr_bands_ref[nb][start:end + 1]) / abs(
                arr_bands_ref[nb][start:end + 1])) / (end + 1 - start) * 100
    else:
        werr = "No nb given"

    return corr, werr



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

from __future__ import division, unicode_literals

"""
This module contains some utility functions and classes that are used in the chemenv package.
"""

__author__ = "David Waroquiers"
__copyright__ = "Copyright 2012, The Materials Project"
__credits__ = "Geoffroy Hautier"
__version__ = "2.0"
__maintainer__ = "David Waroquiers"
__email__ = "david.waroquiers@gmail.com"
__date__ = "Feb 20, 2016"


import numpy as np
from pymatgen.analysis.chemenv.utils.math_utils import power2_inverse_decreasing, power2_decreasing_exp, smootherstep
from pymatgen.analysis.chemenv.utils.math_utils import power2_inverse_power2_decreasing


class RatioFunction(object):

    ALLOWED_FUNCTIONS = {}

    def __init__(self, function, options_dict=None):
        if function not in self.ALLOWED_FUNCTIONS:
            raise ValueError('Function "{}" is not allowed in RatioFunction of '
                             'type "{}"'.format(function, self.__class__.__name__))
        self.eval = object.__getattribute__(self, function)
        self.function = function
        self.setup_parameters(options_dict=options_dict)

    def setup_parameters(self, options_dict):
        function_options = self.ALLOWED_FUNCTIONS[self.function]
        if len(function_options) > 0:
            # Check if there are missing options
            if options_dict is None:
                missing_options = True
            else:
                missing_options = False
                for op in function_options:
                    if op not in options_dict:
                        missing_options = True
                        break
            # If there are missing options, raise an error
            if missing_options:
                if len(function_options) == 1:
                    opts = 'Option "{}"'.format(function_options[0])
                else:
                    opts1 = ', '.join(['"{}"'.format(op) for op in function_options[:-1]])
                    opts = 'Options {}'.format(' and '.join([opts1,
                                                             '"{}"'.format(function_options[-1])]))
                if options_dict is None or len(options_dict) == 0:
                    missing = 'no option was provided.'
                else:
                    optgiven = list(options_dict.keys())
                    if len(options_dict) == 1:
                        missing = 'only "{}" was provided.'.format(optgiven[0])
                    else:
                        missing1 = ', '.join(['"{}"'.format(miss) for miss in optgiven[:-1]])
                        missing = 'only {} were provided.'.format(' and '.join([missing1,
                                                                                '"{}"'.format(optgiven[-1])]))
                raise ValueError('{} should be provided for function "{}" in RatioFunction of '
                                 'type "{}" while {}'.format(opts,
                                                             self.function,
                                                             self.__class__.__name__,
                                                             missing))
            # Setup the options and raise an error if a wrong option is provided
            for key, val in options_dict.items():
                if key not in function_options:
                    raise ValueError('Option "{}" not allowed for function "{}" in RatioFunction of '
                                     'type "{}"'.format(key, self.function, self.__class__.__name__))
                self.__setattr__(key, val)

    def evaluate(self, value):
        return self.eval(value)


class CSMFiniteRatioFunction(RatioFunction):

    ALLOWED_FUNCTIONS = {'power2_decreasing_exp': ['max_csm', 'alpha']}

    def power2_decreasing_exp(self, vals):
        return power2_decreasing_exp(vals, edges=[0.0, self.__dict__['max_csm']], alpha=self.__dict__['alpha'])

    def fractions(self, data):
        if len(data) == 0:
            return None
        total = np.sum([self.eval(dd) for dd in data])
        if total > 0.0:
            return [self.eval(dd) / total for dd in data]
        else:
            return None

    def mean_estimator(self, data):
        if len(data) == 0:
            return None
        elif len(data) == 1:
            return data[0]
        else:
            fractions = self.fractions(data)
            if fractions is None:
                return None
            return np.sum(np.array(fractions) * np.array(data))

    ratios = fractions


class CSMInfiniteRatioFunction(RatioFunction):

    ALLOWED_FUNCTIONS = {'power2_inverse_decreasing': ['max_csm'],
                         'power2_inverse_power2_decreasing': ['max_csm']}

    def power2_inverse_decreasing(self, vals):
        return power2_inverse_decreasing(vals, edges=[0.0, self.__dict__['max_csm']])

    def power2_inverse_power2_decreasing(self, vals):
        return power2_inverse_power2_decreasing(vals, edges=[0.0, self.__dict__['max_csm']])

    def fractions(self, data):
        if len(data) == 0:
            return None
        close_to_zero = np.isclose(data, 0.0, atol=1e-10).tolist()
        nzeros = close_to_zero.count(True)
        if nzeros == 1:
            fractions = [0.0] * len(data)
            fractions[close_to_zero.index(True)] = 1.0
            return fractions
        elif nzeros > 1:
            raise RuntimeError('Should not have more than one continuous symmetry measure with value equal to 0.0')
        else:
            fractions = self.eval(np.array(data))
            total = np.sum(fractions)
            if total > 0.0:
                return fractions / total
            else:
                return None

    def mean_estimator(self, data):
        if len(data) == 0:
            return None
        elif len(data) == 1:
            return data[0]
        else:
            fractions = self.fractions(data)
            if fractions is None:
                return None
            return np.sum(np.array(fractions) * np.array(data))

    ratios = fractions


class DeltaCSMRatioFunction(RatioFunction):

    ALLOWED_FUNCTIONS = {'smootherstep': ['delta_csm_min', 'delta_csm_max']}

    def smootherstep(self, vals):
        return smootherstep(vals, edges=[self.__dict__['delta_csm_min'], self.__dict__['delta_csm_max']])